from __future__ import annotations

import warnings

import attrs
import mitsuba as mi
import numpy as np
import pint
import xarray as xr

from ...attrs import documented, get_doc, parse_docs
from ...constants import EARTH_RADIUS
from ...units import to_quantity
from ...units import unit_context_config as ucc
from ...units import unit_context_kernel as uck
from ...units import unit_registry as ureg
from ..bsdfs import BSDF, LambertianBSDF, OpacityMaskBSDF
from ..core import SceneElement
from ..geometry import PlaneParallelGeometry, SceneGeometry, SphericalShellGeometry
from ..shapes import (
    BufferMeshShape,
    FileMeshShape,
    RectangleShape,
    Shape,
    SphereShape,
    shape_factory,
)
from ._core import Surface


def mesh_from_dem(
    da: xr.DataArray,
    geometry: str | dict | SceneGeometry,
    planet_radius: pint.Quantity | float | None = None,
) -> "mitsuba.Mesh":
    """
    Construct a DEM surface from a data array holding elevation data.

    Parameters
    ----------
    da : DataArray
        Data array with elevation data, indexed either by latitude and longitude
        coordinates or x and y coordinates.

    geometry : .SceneGeometry or dict or str
        Scene geometry configuration. The value is pre-processed by the
        :meth:`.SceneGeometry.convert` converter.

    planet_radius : quantity or float, default: .EARTH_RADIUS
        Planet radius used to convert latitude/longitude to x/y when
        `geometry` is a :class:`.PlaneParallelGeometry` instance.
        This parameter is unused otherwise. If a unitless value is passed, it is
        interpreted using
        :ref:`default config length units <sec-user_guide-unit_guide_user>`.

    Returns
    -------
    mesh : .BufferMeshShape
        A triangulated mesh representing the DEM.

    theta_lim : pint.Quantity
        The limits of the latitude extent of the DEM.

    phi_lim : pint.Quantity
        The limits of the longitude extent of the DEM.

    Notes
    -----
    The ``da`` parameter may use the following formats:

    * with latitude and longitude based coordinates, then named ``"lat"`` and
      ``"lon"``;
    * with x and y length based coordinates, then named ``"x"`` and ``"y"``.

    Coordinate and variable units are specified using the ``units`` xarray
    attributes.
    """
    # Pre-process geometry parameter
    geometry = SceneGeometry.convert(geometry)

    if isinstance(geometry, SphericalShellGeometry) and planet_radius is not None:
        warnings.warn("SphericalShellGeometry overrides the `planet_radius` argument.")

    # Add default units if quantity is unitless
    if planet_radius is not None and not isinstance(planet_radius, pint.Quantity):
        planet_radius = planet_radius * ucc.get("length")
    # Set default if in a plane-parallel setup
    planet_radius = EARTH_RADIUS if planet_radius is None else planet_radius

    if "lat" in da.coords and "lon" in da.coords:
        elevation = to_quantity(da.transpose("lat", "lon"))
        lat = ureg.Quantity(da.lat.values, da["lat"].attrs["units"]).m_as(ureg.deg)
        lon = ureg.Quantity(da.lon.values, da["lon"].attrs["units"]).m_as(ureg.deg)
        vertices = _generate_dem_vertices(lat, lon, elevation.m_as(uck.get("length")))
        faces = _generate_face_indices(len(lat), len(lon))

    elif "x" in da.coords and "y" in da.coords:
        elevation = to_quantity(da.transpose("x", "y"))
        if isinstance(geometry, SphericalShellGeometry):
            planet_radius = geometry.planet_radius

        x = np.rad2deg(
            (ureg.Quantity(da.x.values, da["x"].attrs["units"]) / planet_radius).m
        )
        y = np.rad2deg(
            (ureg.Quantity(da.y.values, da["y"].attrs["units"]) / planet_radius).m
        )

        vertices = _generate_dem_vertices(x, y, elevation.m_as(uck.get("length")))
        faces = _generate_face_indices(len(x), len(y))

    else:
        raise ValueError(
            "Data array coordinates must be either `x/y` or "
            f"`lat/lon`.\nGot: {da.coords}"
        )

    theta_lim, phi_lim = _minmax_coordinates(vertices)
    theta_mean, phi_mean = _mean_coordinates(vertices)
    atmo_bottom = geometry.ground_altitude

    if isinstance(geometry, SphericalShellGeometry):
        trafo = (
            mi.ScalarTransform4f.rotate(axis=[0, 0, 1], angle=-90)
            @ mi.ScalarTransform4f.rotate(axis=[0, 1, 0], angle=-(90 - theta_mean))
            @ mi.ScalarTransform4f.rotate(axis=[0, 0, 1], angle=-phi_mean)
        ).matrix.numpy()[:3, :3]

        vertices_spherical = _transform_vertices_spherical_shell(
            vertices, geometry.planet_radius.m_as(uck.get("length"))
        )
        vertices_new = [trafo.dot(vertex) for vertex in vertices_spherical] * uck.get(
            "length"
        )

        mesh = BufferMeshShape(vertices=vertices_new, faces=faces)

    elif isinstance(geometry, PlaneParallelGeometry):
        vertices_new = _transform_vertices_plane_parallel(
            vertices,
            planet_radius=planet_radius.m_as(uck.get("length")),
            altitude=atmo_bottom.m_as(uck.get("length")),
        ) * uck.get("length")
        mesh = BufferMeshShape(vertices=vertices_new, faces=faces)
    else:
        raise ValueError(
            f"geometry must be PlaneParallelGeometry or SphericalShellGeometry, got {type(geometry)}"
        )

    return mesh, theta_lim * ureg.deg, phi_lim * ureg.deg


def _generate_dem_vertices(x, y, elevation):
    """
    Generate DEM vertex positions as (latitude, longitude, elevation) triplets
    for a plane parallel geometry.

    Parameters
    ----------
    x : array-like
        Points along the latitude axis of the DEM

    y : array-like
        Points along the longitude axis of the DEM

    elevation : array-like
        Elevation data

    Returns
    -------
    vertices : array
        DEM vertices as (lat, lon, elevation) triplets.
    """
    y_gr, x_gr = np.meshgrid(y, x)

    vertices = np.array((x_gr.ravel(), y_gr.ravel(), elevation.ravel())).transpose()
    return vertices


def _vertex_index(x, y, len_y):
    """
    Vertex index for a given row and column of gridded vertex data
    This function handles vectorized index lookup with numpy arrays,
    as well as individual lookup using floats.

    Parameters
    ----------
    x : np.array
        Row index of the vertex grid
    y : np.array
        Column index of the vertex grid
    len_y : float
        Length of the second dimension of the vertex grid

    Returns
    -------
    index : np.array
        Index of the vertex in a flattened structure.
    """
    return x * len_y + y


def _generate_face_indices(len_x, len_y):
    """
    Generate indices for face definition in a mesh with gridded vertex positions

    Parameters
    ----------
    len_x : float
        Length of the row dimension of the vertex grid

    len_y : float
        Length of the column dimension of the vertex grid

    Returns
    -------
    face_indices : np.array
        (n, 3) array defining faces of the triangulated mesh for the DEM.
    """
    x = np.array(range(len_x - 1))
    y = np.array(range(len_y - 1))
    xg, yg = np.meshgrid(x, y)
    vertex_sw = _vertex_index(xg.flatten(), yg.flatten(), len_y)

    xg, yg = np.meshgrid(x + 1, y + 1)
    vertex_ne = _vertex_index(xg.flatten(), yg.flatten(), len_y)

    xg, yg = np.meshgrid(x, y + 1)
    vertex_nw = _vertex_index(xg.flatten(), yg.flatten(), len_y)

    xg, yg = np.meshgrid(x + 1, y)
    vertex_se = _vertex_index(xg.flatten(), yg.flatten(), len_y)

    face_indices_1 = np.array((vertex_ne, vertex_sw, vertex_nw)).transpose()
    face_indices_2 = np.array((vertex_se, vertex_sw, vertex_ne)).transpose()
    face_indices = np.concatenate((face_indices_1, face_indices_2))

    return face_indices


def _transform_vertices_spherical_shell(vertices, planet_radius):
    """
    Convert the (lat, lon, elevation) vertices from the initial vertex generation
    into (x, y, z) values for spherical shell geometries.

    Parameters
    ----------
    vertices : array-like
        List of mesh vertices in (lat, lon, elevation) tuples.

    planet_radius : float
        Planet radius; assumed to be given in kernel units
    """
    theta_r = np.deg2rad(90 - vertices[:, 0])
    phi_r = np.deg2rad(vertices[:, 1])
    x = np.sin(theta_r) * np.cos(phi_r) * (planet_radius + vertices[:, 2])
    y = np.sin(theta_r) * np.sin(phi_r) * (planet_radius + vertices[:, 2])
    z = np.cos(theta_r) * (planet_radius + vertices[:, 2])
    vertices_new = np.array((x, y, z)).transpose()

    return vertices_new


def _transform_vertices_plane_parallel(vertices, planet_radius, altitude):
    """
    Convert the (lat, lon, elevation) vertices from the initial vertex generation
    into (x, y, z) values for plane parallel atmosphere geometries.

    Parameters
    ----------
    vertices : array-like
        List of mesh vertices in (lat, lon, elevation) tuples.

    planet_radius : float
        Planet radius; assumed to be given in kernel units

    altitude : float
        Atmosphere bottom altitude; assumed to be given in kernel units
    """
    theta_mean, phi_mean = _mean_coordinates(vertices)
    phi_local = np.deg2rad(vertices[:, 1] - phi_mean)
    theta_local = np.deg2rad(vertices[:, 0] - theta_mean)

    x = phi_local * planet_radius
    y = theta_local * planet_radius
    z = vertices[:, 2] + altitude

    vertices_new = np.array((x, y, z)).transpose()
    return vertices_new


def _mean_coordinates(vertices):
    """
    Return the mean of the latitude and longitude values in the vertex list.
    The mean is computed as the average of the smallest and largest value in each set.
    """
    (theta_min, theta_max), (phi_min, phi_max) = _minmax_coordinates(vertices)

    phi_mean = ((phi_max + phi_min) / 2.0) % 360
    theta_mean = (theta_max + theta_min) / 2.0
    return theta_mean, phi_mean


def _minmax_coordinates(vertices):
    """
    Return the minimum and maximum values of the latitude and longitude values
    in the vertex list.
    """
    theta_min = np.min(vertices[:, 0])
    theta_max = np.max(vertices[:, 0])
    phi_min = np.min(vertices[:, 1])
    phi_max = np.max(vertices[:, 1])
    return (theta_min, theta_max), (phi_min, phi_max)


def _to_uv(lat_min, lat_max, lon_min, lon_max) -> "mitsuba.ScalarTransform4f":
    """
    Compute the `to_uv` transformation for the opacity mask bitmap. To do this,
    the latitude and longitude extent of the DEM specification are used.

    To avoid intersection between the terrain mesh and the surrounding background
    shape, an opacity mask is attached to the background's BSDF.

    """
    lon_range = lon_max - lon_min
    lon_scale = 120 / lon_range

    lat_range = lat_max - lat_min
    lat_scale = 60 / lat_range

    lon_mean = (lon_max + lon_min) / 2.0
    lon_uv = lon_mean / 360 + 0.5

    lat_mean = (lat_max + lat_min) / 2.0
    lat_uv = 0.5 - (lat_mean / 180)

    return mi.ScalarTransform4f.scale(
        (lon_scale, lat_scale, 1)
    ) @ mi.ScalarTransform4f.translate(
        (-lon_uv + (0.5 / lon_scale), -lat_uv + (0.5 / lat_scale), 0)
    )


@parse_docs
@attrs.define(eq=False, slots=False)
class DEMSurface(Surface):
    """
    DEM Surface [``dem``]

    A mesh based representation of surface DEMs.
    """

    id: str | None = documented(
        attrs.field(
            default="terrain",
            validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        ),
        doc=get_doc(Surface, "id", "doc"),
        type=get_doc(Surface, "id", "type"),
        init_type=get_doc(Surface, "id", "init_type"),
        default='"surface"',
    )

    shape: Shape = documented(
        attrs.field(
            converter=attrs.converters.optional(shape_factory.convert),
            validator=attrs.validators.optional(
                attrs.validators.instance_of((BufferMeshShape, FileMeshShape))
            ),
            kw_only=True,
            default=None,
        ),
        doc="Shape describing the surface.",
        type=".BufferMeshShape or .FileMeshShape or None",
        init_type=".BufferMeshShape or .FileMeshShape or dict",
        default="None",
    )

    shape_background: Shape = documented(
        attrs.field(
            converter=attrs.converters.optional(shape_factory.convert),
            validator=attrs.validators.optional(
                attrs.validators.instance_of((SphereShape, RectangleShape))
            ),
            kw_only=True,
            default=None,
        ),
        doc="Shape describing the background surface.",
        type=".SphereShape or .RectangleShape or None",
        init_type=".SphereShape or .RectangleShape or dict, optional",
        default="None",
    )

    @property
    def _shape_id(self) -> str:
        """
        Mitsuba shape object identifier.
        """
        return f"{self.id}_shape"

    @property
    def _bsdf_id(self) -> str:
        """
        Mitsuba BSDF object identifier.
        """
        return f"{self.id}_bsdf"

    @property
    def _template_shapes(self) -> dict:
        return {}

    @property
    def _params_shapes(self) -> dict:
        return {}

    @property
    def _template_bsdfs(self) -> dict:
        return {}

    @property
    def _params_bsdfs(self) -> dict:
        return {}

    @property
    def objects(self) -> dict[str, SceneElement]:
        # Inherit docstring
        return {
            self._shape_id: self.shape,
            f"{self._shape_id}_background": self.shape_background,
        }

    @classmethod
    def from_mesh(
        cls,
        mesh: BufferMeshShape | FileMeshShape,
        lat: pint.Quantity,
        lon: pint.Quantity,
        id: str = "surface",
        geometry: SceneGeometry = None,
        planet_radius: pint.Quantity = None,
        bsdf: BSDF = None,
    ) -> DEMSurface:
        """
        Construct a DEMSurface from a mesh object and coordinates which specify
        its location.

        Parameters
        ----------
        mesh : .BufferMeshShape or .FileMeshShape
            DEM as a triangulated mesh.

        lat : pint.Quantity
            Limits of the latitude range covered by the DEM.

        lon : pint.Quantity
            Limits of the longitude range covered by the DEM.

        id : str, optional, default: "surface"
            Identifier of the scene element.

        geometry : .SceneGeometry, optional, default: :class:`PlaneParallelGeometry() <.PlaneParallelGeometry>`
            Atmospheric geometry of the scene.

        planet_radius : pint.Quantity, optional, default: .EARTH_RADIUS
            Planet radius. Used only in case of a plane parallel geometry to
            convert between latitude/longitude and x/y coordinates.

        bsdf : .BSDF, optional, default: :class:`LambertianBSDF() <.LambertianBSDF>`
            Scattering model attached to the surface.

        Returns
        -------
        .DEMSurface
        """
        geometry = (
            PlaneParallelGeometry()
            if geometry is None
            else SceneGeometry.convert(geometry)
        )
        bsdf = LambertianBSDF() if bsdf is None else bsdf
        bsdf_id = f"{id}_bsdf"
        mesh = attrs.evolve(mesh, bsdf=bsdf)

        if isinstance(geometry, SphericalShellGeometry):
            if planet_radius is not None:
                warnings.warn(
                    "SphericalShellGeometry overrides the `planet_radius` argument."
                )

            opacity_mask_trafo = _to_uv(*lat.m_as(ureg.deg), *lon.m_as(ureg.deg))
            opacity_array = np.ones((3, 3))
            opacity_array[1, 1] = 0
            opacity_bitmap = mi.Bitmap(opacity_array)

            opacity_bsdf = OpacityMaskBSDF(
                id=f"{bsdf_id}_shape_background",
                nested_bsdf=bsdf,
                opacity_bitmap=opacity_bitmap,
                uv_trafo=opacity_mask_trafo,
            )

            # The 'hole' in the background surface is created at the equator:
            # we rotate it to
            trafo = (
                mi.ScalarTransform4f.rotate(axis=[0, 0, 1], angle=90)
                @ mi.ScalarTransform4f.rotate(
                    axis=[0, 1, 0], angle=(90 - np.mean(lat.m_as(ureg.deg)))
                )
                @ mi.ScalarTransform4f.rotate(
                    axis=[0, 0, 1], angle=-np.mean(lon.m_as(ureg.deg))
                )
            )

            surface_background = SphereShape(
                id=f"{id}_shape_background",
                center=[0.0, 0.0, 0.0] * geometry.planet_radius.units,
                radius=geometry.planet_radius + geometry.ground_altitude,
                bsdf=opacity_bsdf,
                to_world=trafo,
            )

        elif isinstance(geometry, PlaneParallelGeometry):
            planet_radius = EARTH_RADIUS if planet_radius is None else planet_radius

            lat_length = (lat[1] - lat[0]).m_as(ureg.rad) * planet_radius
            lat_scale = (geometry.width / (lat_length * 3)).magnitude
            lon_length = (lon[1] - lon[0]).m_as(ureg.rad) * planet_radius
            lon_scale = (geometry.width / (lon_length * 3)).magnitude
            opacity_mask_trafo = mi.ScalarTransform4f.scale(
                (lon_scale, lat_scale, 1)
            ) @ mi.ScalarTransform4f.translate(
                (-0.5 + (0.5 / lon_scale), -0.5 + (0.5 / lat_scale), 0)
            )
            opacity_array = np.ones((3, 3))
            opacity_array[1, 1] = 0
            opacity_bsdf = OpacityMaskBSDF(
                nested_bsdf=bsdf,
                opacity_bitmap=opacity_array,
                uv_trafo=opacity_mask_trafo,
            )
            surface_background = RectangleShape(
                id=f"{id}_background",
                edges=geometry.width,
                center=[0.0, 0.0, geometry.ground_altitude.m]
                * geometry.ground_altitude.units,
                normal=np.array([0, 0, 1]),
                up=np.array([0, 1, 0]),
                bsdf=opacity_bsdf,
            )

        else:
            raise TypeError(f"Scene geometry is not recognized. {geometry}")

        return cls(shape=mesh, shape_background=surface_background, id=id)



from __future__ import annotations

import attrs
import numpy as np
import pint
import pinttr

from ._core import Measure
from ...attrs import documented, parse_docs
from ...units import symbol
from ...units import unit_context_config as ucc
from ...units import unit_context_kernel as uck
from ...units import unit_registry as ureg


@parse_docs
@attrs.define(eq=False, slots=False)
class MultiRadiancemeterMeasure(Measure):
    """
    Radiance meter array measure scene element [``mradiancemeter``,
    ``multi_radiancemeter``].

    This measure scene element is a thin wrapper around the ``mradiancemeter``
    sensor kernel plugin. It records the incident power per unit area per unit
    solid angle along a number of rays defined by its ``origins`` and
    ``directions`` parameters.
    """

    # --------------------------------------------------------------------------
    #                           Fields and properties
    # --------------------------------------------------------------------------

    origins: pint.Quantity = documented(
        pinttr.field(
            default=ureg.Quantity([[0.0, 0.0, 0.0]], ureg.m),
            units=ucc.deferred("length"),
        ),
        doc="A sequence of points specifying radiance meter array positions.\n"
        "\n"
        "Unit-enabled field (default: ucc['length']).",
        type="quantity",
        init_type="array-like",
        default="[[0, 0, 0]] m",
    )

    directions: np.ndarray = documented(
        attrs.field(
            default=np.array([[0.0, 0.0, 1.0]]),
            converter=np.array,
        ),
        doc="A sequence of 3-vectors specifying radiance meter array directions.",
        type="quantity",
        init_type="array-like",
        default="[[0, 0, 1]]",
    )

    @directions.validator
    @origins.validator
    def _target_origin_validator(self, attribute, value):
        if value.shape[1] != 3:
            raise ValueError(
                f"While initializing {attribute}: "
                f"Expected shape (N, 3), got {value.shape}"
            )

        if not self.origins.shape == self.directions.shape:
            raise ValueError(
                f"While initializing {attribute}: "
                f"Origin and direction arrays must have the same shape, "
                f"got origins.shape = {self.origins.shape}, "
                f"directions.shape = {self.directions.shape}"
            )

    @property
    def film_resolution(self) -> tuple[int, int]:
        return (self.origins.shape[0], 1)

    # --------------------------------------------------------------------------
    #                       Kernel dictionary generation
    # --------------------------------------------------------------------------

    @property
    def kernel_type(self) -> str:
        # Inherit docstring
        return "mradiancemeter"

    @property
    def template(self) -> dict:
        # Inherit docstring
        result = super().template

        origins = self.origins.m_as(uck.get("length"))
        result["origins"] = ",".join(map(str, origins.ravel(order="C")))
        directions = self.directions
        result["directions"] = ",".join(map(str, directions.ravel(order="C")))

        return result

    # --------------------------------------------------------------------------
    #                        Post-processing information
    # --------------------------------------------------------------------------

    @property
    def var(self) -> tuple[str, dict]:
        return "radiance", {
            "standard_name": "radiance",
            "long_name": "radiance",
            "units": symbol(uck.get("radiance")),
        }



from __future__ import annotations

import typing as t
from collections import OrderedDict
from pathlib import Path

import attrs

from ._core import DataStore
from ..attrs import documented, parse_docs
from ..exceptions import DataError
from ..typing import PathLike


@parse_docs
@attrs.define
class MultiDataStore(DataStore):
    """
    Chain requests on multiple data stores.

    Calls to the :meth:`~.MultiDataStore.fetch` method are successively redirected
    to each referenced data store. The first successful request is served.
    """

    stores: OrderedDict = documented(
        attrs.field(factory=OrderedDict, converter=OrderedDict),
        type="collections.OrderedDict",
        init_type="mapping",
        default="{}",
        doc="Data stores which will be queried successively.",
    )

    def __getitem__(self, item):
        return self.stores[item]

    @property
    def base_url(self) -> str:
        """
        Raises :class:`NotImplementedError` (this data store has no target
        location).
        """
        raise NotImplementedError

    @property
    def registry(self) -> dict:
        """
        Raises :class:`NotImplementedError` (this data store has no registry).
        """
        raise NotImplementedError

    def registry_files(
        self, filter: t.Callable[[t.Any], bool] | None = None
    ) -> list[str]:
        """
        Returns an empty list (this data store has no registry).
        """
        return []

    def fetch(self, filename: PathLike, **kwargs) -> Path:
        # Inherit docstring
        # No kwargs are actually accepted
        if kwargs:
            keyword = next(iter(kwargs.keys()))
            raise TypeError(f"fetch() got an unexpected keyword argument '{keyword}'")

        # Try and serve data
        for _, store in self.stores.items():
            try:
                return store.fetch(filename)
            except DataError:
                continue

        raise DataError(f"file '{filename}' could not be served")



"""
A collection of tools which don't really fit anywhere else.
"""

from __future__ import annotations

import functools
import inspect
import re
import typing as t
from collections import OrderedDict
from numbers import Number

import numpy as np
import numpy.typing as npt
import pint
import xarray as xr


class cache_by_id:
    """
    Cache the result of a function based on the ID of its arguments.

    This decorator caches the value returned by the function it wraps in order
    to avoid unnecessary execution upon repeated calls with the same arguments.

    Warnings
    --------
    The main difference with
    :func:`functools.lru_cache(maxsize=1) <functools.lru_cache>` is that the
    cache is referenced by positional argument IDs instead of hashes.
    Therefore, this decorator can be used with NumPy arrays; but it's also
    unsafe, because mutating an argument won't trigger a recompute, while it
    actually shoud! **Use with great care!**

    Notes
    -----
    * Meant to be used as a decorator.
    * The wrapped function may only have positional arguments.
    * Works with functions and methods.

    Examples
    --------
    >>> @cache_by_id
    ... def f(x, y):
    ...     print("Calling f")
    ...     return x, y
    >>> f(1, 2)
    Calling f
    (1, 2)
    >>> f(1, 2)
    (1, 2)
    >>> f(1, 1)
    Calling f
    (1, 1)
    >>> f(1, 1)
    (1, 1)
    """

    def __init__(self, func):
        functools.update_wrapper(self, func)
        self.func = func
        self._cached_value = None
        self._cached_index = None

    def __call__(self, *args):
        index = tuple(id(arg) for arg in args)

        if index != self._cached_index:
            self._cached_index = index
            self._cached_value = self.func(*args)

        return self._cached_value

    def __get__(self, instance, owner):
        # See https://stackoverflow.com/questions/30104047 for full explanation
        return functools.partial(self.__call__, instance)


class LoggingContext(object):
    """
    This context manager allows for a temporary override of logger settings.
    """

    # from https://docs.python.org/3/howto/logging-cookbook.html#using-a-context-manager-for-selective-logging
    def __init__(self, logger, level=None, handler=None, close=True):
        self.logger = logger
        self.level = level
        self.handler = handler
        self.close = close

    def __enter__(self):
        if self.level is not None:
            self.old_level = self.logger.level
            self.logger.setLevel(self.level)
        if self.handler:
            self.logger.addHandler(self.handler)

    def __exit__(self, et, ev, tb):
        if self.level is not None:
            self.logger.setLevel(self.old_level)
        if self.handler:
            self.logger.removeHandler(self.handler)
        if self.handler and self.close:
            self.handler.close()
        # implicit return of None => don't swallow exceptions


class Singleton(type):
    """
    A simple singleton implementation. See [1]_ for details.

    References
    -------
    .. [1] `Creating a singleton in Python on Stack Overflow <https://stackoverflow.com/questions/6760685/creating-a-singleton-in-python>`_.

    Examples
    --------

    .. testsetup:: singleton

       from eradiate.util.misc import Singleton

    .. doctest:: singleton

       >>> class MySingleton(metaclass=Singleton): ...
       >>> my_singleton1 = MySingleton()
       >>> my_singleton2 = MySingleton()
       >>> my_singleton1 is my_singleton2
       True

    .. testcleanup:: singleton

       del Singleton
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


def camel_to_snake(name):
    # from https://stackoverflow.com/questions/1175208/elegant-python-function-to-convert-camelcase-to-snake-case
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()


def deduplicate(value: t.Sequence, preserve_order: bool = True) -> list:
    """
    Remove duplicates from a sequence.

    Parameters
    ---------
    value : sequence
        Sequence to remove duplicates from.

    preserve_order : bool, optional, default: True
        If ``True``, preserve item ordering. The first occurrence of duplicated
        items is kept. Setting to ``False`` may slightly improve performance.

    Returns
    -------
    list
        List of values with duplicates removed.
    """

    if preserve_order:
        return list(OrderedDict.fromkeys(value))

    else:
        return list(set(value))


def deduplicate_sorted(value: t.Sequence, cmp: t.Callable | None = None) -> list:
    if cmp is None:
        cmp = lambda x, y: x == y  # noqa: E731

    result = [value[0]]

    for i in range(1, len(value)):
        if not cmp(value[i], value[i - 1]):
            result.append(value[i])

    return result


def flatten(d: t.Mapping, sep: str = ".", name: str = "") -> dict:
    """
    Flatten a nested dictionary.

    Parameters
    ----------
    d : dict
        Dictionary to be flattened.

    name : str, optional, default: ""
        Path to the parent dictionary. By default, no parent name is defined.

    sep : str, optional, default: "."
        Flattened dict key separator.

    Returns
    -------
    dict
        A flattened copy of `d`.

    See Also
    --------
    :func:`.nest`, :func:`.set_nested`
    """
    result = {}

    for k, v in d.items():
        full_key = k if not name else f"{name}{sep}{k}"
        if isinstance(v, dict):
            result.update(flatten(v, sep=sep, name=full_key))
        else:
            result[full_key] = v

    return result


def fullname(obj: t.Any) -> str:
    """
    Get the fully qualified name of `obj`. Aliases will be dereferenced.
    """
    cls = get_class_that_defined_method(obj)

    if cls is None:
        return f"{obj.__module__}.{obj.__qualname__}"

    # else: # (it's a method)
    return f"{cls.__module__}.{obj.__qualname__}"


def get_class_that_defined_method(meth: t.Any) -> type:
    """
    Get the class which defined a method, if relevant. Otherwise, return
    ``None``.
    """
    # See https://stackoverflow.com/questions/3589311/get-defining-class-of-unbound-method-object-in-python-3/25959545#25959545
    if isinstance(meth, functools.partial):
        return get_class_that_defined_method(meth.func)

    if inspect.ismethod(meth) or (
        inspect.isbuiltin(meth)
        and getattr(meth, "__self__", None) is not None
        and getattr(meth.__self__, "__class__", None)
    ):
        for cls in inspect.getmro(meth.__self__.__class__):
            if meth.__name__ in cls.__dict__:
                return cls
        meth = getattr(meth, "__func__", meth)  # fallback to __qualname__ parsing

    if inspect.isfunction(meth):
        cls = getattr(
            inspect.getmodule(meth),
            meth.__qualname__.split(".<locals>", 1)[0].rsplit(".", 1)[0],
            None,
        )
        if isinstance(cls, type):
            return cls

    return getattr(meth, "__objclass__", None)  # handle special descriptor objects


def is_vector3(value: t.Any):
    """
    Check if value can be interpreted as a 3-vector.

    Parameters
    ----------
    value
        Value to be checked.

    Returns
    -------
    bool
        ``True`` if a value can be interpreted as a 3-vector.
    """

    if isinstance(value, pint.Quantity):
        return is_vector3(value.magnitude)

    return (
        (
            isinstance(value, np.ndarray)
            or (isinstance(value, t.Sequence) and not isinstance(value, str))
        )
        and len(value) == 3
        and all(map(lambda x: isinstance(x, Number), value))
    )


def natsort_alphanum_key(x):
    """
    Simple sort key natural order for string sorting. See [2]_ for details.

    See Also
    --------
    `Sorting HOWTO <https://docs.python.org/3/howto/sorting.html>`_

    References
    ----------
    .. [2] `Natural sorting on Stack Overflow <https://stackoverflow.com/a/11150413/3645374>`_.
    """
    return tuple(
        map(
            lambda text: int(text) if text.isdigit() else text.lower(),
            re.split("([0-9]+)", x),
        )
    )


def natsorted(l):  # noqa
    """
    Sort a list of strings with natural ordering.

    Parameters
    ----------
    l : iterable
        List to sort.

    Returns
    -------
    list
        List sorted using :func:`natsort_alphanum_key`.
    """
    return sorted(l, key=natsort_alphanum_key)


def nest(d: t.Mapping, sep: str = ".") -> dict:
    """
    Turn a flat dictionary into a nested dictionary.

    Parameters
    ----------
    d : dict
        Dictionary to be unflattened.

    sep : str, optional, default: "."
        Flattened dict key separator.

    Returns
    -------
    dict
        A nested copy of `d`.

    See Also
    --------
    :func:`.flatten`, :func:`.set_nested`
    """
    result = {}

    for key, value in d.items():
        set_nested(result, key, value, sep)

    return result


def onedict_value(d: t.Mapping) -> t.Any:
    """
    Get the value of a single-entry dictionary.

    Parameters
    ----------
    d : mapping
        A single-entry mapping.

    Returns
    -------
    object
        Unwrapped value.

    Raises
    ------
    ValueError
        If ``d`` has more than a single element.

    Notes
    -----
    This function is basically ``next(iter(d.values()))`` with a safeguard.

    Examples
    --------

    .. testsetup:: onedict_value

       from eradiate.util.misc import onedict_value

    .. doctest:: onedict_value

       >>> onedict_value({"foo": "bar"})
       'bar'

    .. testcleanup:: onedict_value

       del onedict_value
    """

    if len(d) != 1:
        raise ValueError(f"dictionary has wrong length (expected 1, got {len(d)}")

    return next(iter(d.values()))


def round_to_multiple(number, multiple, direction="nearest"):
    if direction == "nearest":
        return multiple * round(number / multiple)
    elif direction == "up":
        return multiple * np.ceil(number / multiple)
    elif direction == "down":
        return multiple * np.floor(number / multiple)
    else:
        return multiple * round(number / multiple)


def set_nested(d: t.Mapping, path: str, value: t.Any, sep: str = ".") -> None:
    """
    Set values in a nested dictionary using a flat path.

    Parameters
    ----------
    d : dict
        Dictionary to operate on.

    path : str
        Path to the value to be set.

    value
        Value to which `path` is to be set.

    sep : str, optional, default: "."
        Separator used to decompose `path`.

    See Also
    --------
    :func:`.flatten`, :func:`.nest`
    """
    *path, last = path.split(sep)
    for bit in path:
        d = d.setdefault(bit, {})
    d[last] = value


def str_summary_numpy(x):
    with np.printoptions(
        threshold=4, edgeitems=2, formatter={"float_kind": lambda x: f"{x:g}"}
    ):
        shape_str = ",".join(map(str, x.shape))
        prefix = f"array<{shape_str}>("
        array_str = f"{x}"

        # Indent repr if it is multiline
        split = array_str.split("\n")
        if len(split) > 1:
            array_str = ("\n" + " " * len(prefix)).join(split)

        return f"{prefix}{array_str})"


@functools.singledispatch
def summary_repr(value):
    """
    Return a summarized repr for `value`.
    """
    return repr(value)


@summary_repr.register
def _(ds: xr.Dataset):
    extra_info = {}

    try:
        extra_info["source"] = repr(ds.encoding["source"])
    except KeyError:
        pass

    desc = ", ".join([f"{key}={value}" for key, value in extra_info.items()])

    if desc:
        desc = " | " + desc

    return f"<xarray.Dataset{desc}>"


@summary_repr.register
def _(da: xr.DataArray):
    extra_info = {}

    try:
        extra_info["name"] = repr(da.name)
    except AttributeError:
        pass

    extra_info["dims"] = repr(list(da.dims))

    try:
        extra_info["source"] = repr(da.encoding["source"])
    except KeyError:
        pass

    desc = ", ".join([f"{key}={value}" for key, value in extra_info.items()])

    if desc:
        desc = " | " + desc

    return f"<xarray.DataArray{desc}>"


def summary_repr_vector(a: np.ndarray, edgeitems: int = 4):
    """
    Return a brief summary representation of a Numpy vector.
    """
    size = len(a)
    if size > edgeitems * 2 + 1:
        return (
            f"[{np.array2string(a[:edgeitems]).strip('[]')}"
            " ... "
            f"{np.array2string(a[size-edgeitems:]).strip('[]')}]"
        )
    else:
        return np.array2string(a)


def find_runs(
    x: npt.ArrayLike,
) -> tuple[npt.ArrayLike, npt.ArrayLike, npt.ArrayLike]:
    """
    Find runs of consecutive items in an array.

    Parameters
    ----------
    x : array-like
        Input array.

    Returns
    -------
    tuple(array-like, array-like, array-like)
        Run values, run starts, run lengths.

    Notes
    -----
    Credit: Alistair Miles
    Source: https://gist.github.com/alimanfoo/c5977e87111abe8127453b21204c1065
    """

    # ensure array
    x = np.asanyarray(x)
    if x.ndim != 1:
        raise ValueError("only 1D array supported")
    n = x.shape[0]

    # handle empty array
    if n == 0:
        return np.array([]), np.array([]), np.array([])

    else:
        # find run starts
        loc_run_start = np.empty(n, dtype=bool)
        loc_run_start[0] = True
        np.not_equal(x[:-1], x[1:], out=loc_run_start[1:])
        run_starts = np.nonzero(loc_run_start)[0]

        # find run values
        run_values = x[loc_run_start]

        # find run lengths
        run_lengths = np.diff(np.append(run_starts, n))

        return run_values, run_starts, run_lengths



from . import gridvolume as gridvolume
from . import logging as logging
from . import transform as transform
from ._bitmap import bitmap_to_dataarray as bitmap_to_dataarray
from ._bitmap import bitmap_to_dataset as bitmap_to_dataset
from ._bsdf import eval_bsdf as eval_bsdf
from ._kernel_dict import InitParameter as InitParameter
from ._kernel_dict import KernelDictTemplate as KernelDictTemplate
from ._kernel_dict import UpdateMapTemplate as UpdateMapTemplate
from ._kernel_dict import UpdateParameter as UpdateParameter
from ._render import MitsubaObjectWrapper as MitsubaObjectWrapper
from ._render import TypeIdLookupStrategy as TypeIdLookupStrategy
from ._render import mi_render as mi_render
from ._render import mi_traverse as mi_traverse
from ._versions import check_kernel as check_kernel
from .gridvolume import read_binary_grid3d as read_binary_grid3d
from .gridvolume import write_binary_grid3d as write_binary_grid3d
from .logging import install_logging as install_logging
from .transform import map_cube as map_cube
from .transform import map_unit_cube as map_unit_cube



from __future__ import annotations

import typing as t

import attrs

from .attrs import documented, parse_docs
from .spectral import SpectralIndex

# ------------------------------------------------------------------------------
#                                      ABC
# ------------------------------------------------------------------------------


@attrs.define
class Context:
    """Abstract base class for all context data structures."""

    def evolve(self, **changes):
        """
        Create a copy of self with changes applied.

        Parameters
        ----------
        **changes
            Keyword changes in the new copy.

        Returns
        -------
        <same type as self>
            A copy of self with ``changes`` incorporated.
        """
        return attrs.evolve(self, **changes)


# ------------------------------------------------------------------------------
#                         Kernel dictionary contexts
# ------------------------------------------------------------------------------


@parse_docs
@attrs.define
class KernelContext(Context):
    """
    Kernel evaluation context data structure. This class is used *e.g.* to store
    information about the spectral configuration to apply when generating kernel
    dictionaries or update parameter maps associated with a
    :class:`.SceneElement` instance.
    """

    si: SpectralIndex = documented(
        attrs.field(
            factory=SpectralIndex.new,
            converter=SpectralIndex.convert,
            validator=attrs.validators.instance_of(SpectralIndex),
        ),
        doc="Spectral index (used to evaluate quantities with any degree "
        "or kind of dependency vs spectrally varying quantities).",
        type=":class:`.SpectralIndex`",
        init_type=":class:`.SpectralIndex` or dict",
        default=":meth:`SpectralIndex.new() <.SpectralIndex.new>`",
    )

    kwargs: dict[str, t.Any] = documented(
        attrs.field(factory=dict),
        doc="Object-specific parameter overrides.",
        type="dict",
        default="{}",
    )

    @property
    def index_formatted(self) -> str:
        return self.si.formatted_repr


# ------------------------------------------------------------------------------
#                         Context Generator
# ------------------------------------------------------------------------------


class MultiGenerator:
    """
    This generator aggregates several generators and makes sure that items that
    have already been served are not repeated.
    """

    def __init__(self, generators):
        self.generators = generators
        self._i_generator = 0
        self._current_iterator = iter(self.generators[self._i_generator])
        self._visited = set()

    def __iter__(self):
        return self

    def __next__(self):
        try:
            result = next(self._current_iterator)
            if result not in self._visited:
                self._visited.add(result)
                return result
            else:
                return self.__next__()
        except StopIteration:
            if self._i_generator >= len(self.generators) - 1:
                raise
            else:
                self._i_generator += 1
                self._current_iterator = iter(self.generators[self._i_generator])
                return self.__next__()



from __future__ import annotations

import logging
import typing as t

import attrs
import numpy as np
import pint
import pinttr
import xarray as xr

from .index import MonoSpectralIndex
from .spectral_set import SpectralSet
from ..attrs import documented, parse_docs
from ..constants import SPECTRAL_RANGE_MAX, SPECTRAL_RANGE_MIN
from ..radprops import MonoAbsorptionDatabase
from ..units import to_quantity
from ..units import unit_context_config as ucc
from ..units import unit_registry as ureg

logger = logging.getLogger(__name__)


@parse_docs
@attrs.define(eq=False, frozen=True, slots=True)
class WavelengthSet(SpectralSet):
    """
    A data class representing a wavelength set used in monochromatic modes.
    """

    wavelengths: pint.Quantity = documented(
        pinttr.field(
            units=ucc.deferred("length"),
            on_setattr=None,
            converter=np.atleast_1d,
        ),
        doc="Wavelengths.",
        type="quantity",
        init_type="quantity or array-like or float",
    )

    def spectral_indices(self) -> t.Generator[MonoSpectralIndex]:
        for w in self.wavelengths:
            yield MonoSpectralIndex(w=w)

    @classmethod
    def from_absorption_database(cls, abs_db: MonoAbsorptionDatabase) -> WavelengthSet:
        """
        Create a wavelength set from an absorption dataset.

        Parameters
        ----------
        abs_db : .MonoAbsorptionDatabase
            Absorption dataset.

        Returns
        -------
        .WavelengthSet
        """
        wavelengths = (
            np.sort(abs_db.spectral_coverage.index.get_level_values(1).values) * ureg.nm
        )
        return cls(wavelengths=wavelengths)

    @classmethod
    def arange(
        cls,
        start: pint.Quantity,
        stop: pint.Quantity,
        step: pint.Quantity,
    ) -> WavelengthSet:
        """
        Create a wavelength set from an array of wavelengths.

        Parameters
        ----------
        start : quantity
            First wavelength.

        stop : quantity
            Last wavelength.

        step : quantity
            Wavelength step.

        Returns
        -------
        :class:`.WavelengthSet`
            Generated wavelength set.
        """
        wunits = ucc.get("wavelength")
        return cls(
            wavelengths=np.arange(
                start.m_as(wunits),
                stop.m_as(wunits),
                step.m_as(wunits),
            )
            * wunits
        )

    @classmethod
    def from_srf(
        cls,
        srf: xr.Dataset,
        step: pint.Quantity = 10.0 * ureg.nm,
    ) -> WavelengthSet:
        """
        Generate a wavelength set with linearly spaced bins, that covers the
        spectral range of a spectral response function.

        Parameters
        ----------
        srf: Dataset
            Spectral response function dataset.

        step : quantity
            Wavelength step.
        """
        wavelengths = to_quantity(srf.w)
        wmin = wavelengths.min()
        wmax = wavelengths.max()

        return cls.arange(
            start=wmin - step,
            stop=wmax + step,
            step=step,
        )

    @classmethod
    def default(cls) -> WavelengthSet:
        """
        Generate a default wavelength set, which covers Eradiate's default
        spectral range with 1 nm spacing.
        """
        dw = 1.0 * ureg.nm
        return cls.arange(
            start=SPECTRAL_RANGE_MIN,
            stop=SPECTRAL_RANGE_MAX + dw,
            step=dw,
        )



"""
Command-line interface to the :mod:`~eradiate.srf_tools` module.
"""

from pathlib import Path
from typing import Optional

import pint
import typer
import xarray as xr
from rich.console import Console
from typing_extensions import Annotated

from eradiate import srf_tools
from eradiate import unit_registry as ureg

app = typer.Typer()
console = Console(color_system=None)


@app.callback()
def main():
    """
    Spectral response function filtering utility.
    """
    pass


@app.command()
def trim(
    filename: Annotated[
        Path,
        typer.Argument(
            # exists=True,
            # help="path to the spectral response function data set to process.",
        ),
    ],
    output: Annotated[
        Path,
        typer.Argument(help="File where to write the filtered data set."),
    ],
    verbose: Annotated[
        bool,
        typer.Option("-v", "--verbose", help="Display filtering summary."),
    ] = False,
    show_plot: Annotated[
        bool,
        typer.Option("-s", "--show-plot", help="Show plot of the filtered region."),
    ] = False,
    dry_run: Annotated[
        bool,
        typer.Option("-d", "--dry-run", help="Do not write filtered data to disk."),
    ] = False,
    interactive: Annotated[
        bool,
        typer.Option(
            "-i",
            "--interactive",
            help="Prompt user to proceed to saving the filtered dataset.",
        ),
    ] = False,
):
    """
    Trim a spectral response function.
    Remove all-except-last leading zeros and all-except-first trailing zeros.
    """

    srf_tools.trim_and_save(
        srf=filename,
        path=output,
        verbose=verbose,
        show_plot=show_plot,
        dry_run=dry_run,
        interactive=interactive,
    )


def text_input_to_quantity(
    value: Optional[str], default_units: str = "nm"
) -> Optional[pint.Quantity]:
    """
    Converts text input to wavelength quantity.

    Parameters
    ----------
    value: str or None
        Value to convert to quantity.

    default_units: str, optional
        If ``value`` converts to a dimensionless quantity, use these units.

    Returns
    -------
    quantity, optional
        Converted quantity.
    """
    if value is None:
        return None
    else:
        # try to parse value into wavelength quantity
        parsed = ureg(value)
        if isinstance(parsed, pint.Quantity):
            return parsed
        else:  # float
            return ureg.Quantity(parsed, default_units)


@app.command()
def filter(
    filename: Annotated[
        Path,
        typer.Argument(
            exists=True,
            help=" path to the spectral response function data to filter.",
        ),
    ],
    output: Annotated[
        Path,
        typer.Argument(
            help="Path where to write the filtered data.",
        ),
    ],
    trim: Annotated[
        bool,
        typer.Option(
            help="Trim the data set prior to filtering.",
        ),
    ] = True,
    verbose: Annotated[
        bool,
        typer.Option(
            "--verbose",
            "-v",
            help="Display filtering summary.",
        ),
    ] = False,
    dry_run: Annotated[
        bool,
        typer.Option(
            "-d",
            "--dry-run",
            help="Do not write filtered data to file.",
        ),
    ] = False,
    interactive: Annotated[
        bool,
        typer.Option(
            "-i",
            "--interactive",
            help="Prompt user to proceed to saving the filtered dataset.",
        ),
    ] = False,
    show_plot: Annotated[
        bool,
        typer.Option(
            "-s",
            "--show-plot",
            help="Show plot of the filtered region.",
        ),
    ] = False,
    threshold: Annotated[
        Optional[float],
        typer.Option(
            "-t",
            "--threshold",
            help="Data points where response is less then or equal to this "
            "value are dropped.",
        ),
    ] = None,
    wmin: Annotated[
        Optional[float],
        typer.Option(
            "--wmin",
            "-w",
            help="Lower wavelength value in nm.",
        ),
    ] = None,
    wmax: Annotated[
        Optional[float],
        typer.Option(
            "--wmax",
            "-W",
            help="Upper wavelength value in nm.",
        ),
    ] = None,
    percentage: Annotated[
        Optional[float],
        typer.Option(
            "-p",
            "--percentage",
            help="Data points that do not contribute to this percentage of the "
            "integrated spectral response are dropped",
        ),
    ] = None,
):
    """
    Filter a spectral response function data set.
    """
    # input conversion
    input_path = Path(filename).absolute()
    srf = xr.load_dataset(input_path)
    wmin = text_input_to_quantity(value=wmin)
    wmax = text_input_to_quantity(value=wmax)

    # filter
    srf_tools.filter_srf(
        srf=srf,
        path=output,
        trim_prior=trim,
        verbose=verbose,
        show_plot=show_plot,
        dry_run=dry_run,
        interactive=interactive,
        threshold=threshold,
        wmin=wmin,
        wmax=wmax,
        percentage=percentage,
    )



from __future__ import annotations

__all__ = [
    "symbol",
    "to_quantity",
    "units_compatible",
    "unit_context_config",
    "unit_context_kernel",
    "unit_registry",
    "PhysicalQuantity",
]

import enum
import importlib
import logging
import typing as t
from functools import lru_cache
from importlib.resources import files

import pint
import pinttr
import xarray
from pinttr.exceptions import UnitsError
from pinttr.util import units_compatible

logger = logging.getLogger(__name__)


# -- Global data members -------------------------------------------------------

# Joseki being a dependency of Eradiate, and given that it uses the application
# registry, we import its definitions early
importlib.import_module("joseki.units")

#: Unit registry common to all Eradiate components. All units used in Eradiate
#: must be created using this registry. Aliased in :mod:`eradiate`.
unit_registry = pint.get_application_registry()


def _parse_definitions(path):
    # Parse a unit definition file (i.e. strip it from line comments and empty
    # lines)
    definitions = []
    with open(path, "r") as f:
        for line in f:
            line = line.strip()

            if line.startswith("#") or not line:  # Skip comments and empty lines
                continue
            else:
                definitions.append(line)

    return definitions


def _load_definitions(ureg: pint.UnitRegistry, definitions: list[str]) -> None:
    # Add extra definitions, possibly skipping those that already exist. This is
    # a very simple wrapper around Pint's usual unit definition logic.

    for definition in definitions:
        _unit_name, _unit_definition = list(
            map(lambda x: x.strip(), definition.split("="))
        )[0:2]

        if _unit_name in ureg:
            if 1.0 * ureg(_unit_definition) == 1.0 * ureg(_unit_name):
                # Definitions are identical: skip
                # Note: we don't check symbols or aliases, might cause bugs in
                # very rare cases
                continue
            else:
                # Definitions are different: attempt a redefinition and let
                # user-controlled conflict handling policy apply
                ureg.define(definition)
        else:
            # Attempt definition
            ureg.define(definition)


_load_definitions(unit_registry, _parse_definitions(files("eradiate") / "units.txt"))


class PhysicalQuantity(enum.Enum):
    """An enumeration defining physical quantities known to Eradiate."""

    ALBEDO = "albedo"
    ANGLE = "angle"
    COLLISION_COEFFICIENT = "collision_coefficient"
    DIMENSIONLESS = "dimensionless"
    INTENSITY = "intensity"
    IRRADIANCE = "irradiance"
    LENGTH = "length"
    MASS = "mass"
    RADIANCE = "radiance"
    REFLECTANCE = "reflectance"
    SPEED = "speed"
    TIME = "time"
    TRANSMITTANCE = "transmittance"
    WAVELENGTH = "wavelength"
    WAVENUMBER = "wavenumber"

    @classmethod
    @lru_cache(maxsize=32)
    def spectrum(cls):
        """
        Return a tuple containing a subset of :class:`PhysicalQuantity`
        members suitable for :class:`.Spectrum` initialisation. This function
        caches its results for improved efficiency.
        """
        return (
            cls.ALBEDO,
            cls.COLLISION_COEFFICIENT,
            cls.DIMENSIONLESS,
            cls.INTENSITY,
            cls.IRRADIANCE,
            cls.RADIANCE,
            cls.REFLECTANCE,
            cls.TRANSMITTANCE,
            cls.ANGLE,
        )


def _make_unit_context():
    uctx = pinttr.UnitContext(
        interpret_str=True, ureg=unit_registry, key_converter=PhysicalQuantity
    )

    # fmt: off
    for key, value in {
        # We allow for dimensionless quantities
        PhysicalQuantity.DIMENSIONLESS: pinttr.UnitGenerator(unit_registry.dimensionless),  # noqa: E501
        # Basic quantities must be named after their SI name
        # https://en.wikipedia.org/wiki/International_System_of_Units
        PhysicalQuantity.LENGTH: pinttr.UnitGenerator(unit_registry.m),
        PhysicalQuantity.TIME: pinttr.UnitGenerator(unit_registry.s),
        PhysicalQuantity.MASS: pinttr.UnitGenerator(unit_registry.kg),
        # Derived quantity names are more flexible
        PhysicalQuantity.ALBEDO: pinttr.UnitGenerator(unit_registry.dimensionless),
        PhysicalQuantity.ANGLE: pinttr.UnitGenerator(unit_registry.deg),
        PhysicalQuantity.REFLECTANCE: pinttr.UnitGenerator(unit_registry.dimensionless),
        PhysicalQuantity.TRANSMITTANCE: pinttr.UnitGenerator(unit_registry.dimensionless),  # noqa: E501
        PhysicalQuantity.WAVELENGTH: pinttr.UnitGenerator(unit_registry.nm),
        PhysicalQuantity.WAVENUMBER: pinttr.UnitGenerator(unit_registry.cm ** -1),
    }.items():
        uctx.register(key, value)
    # fmt: on

    # The following quantities will update automatically based on their parent units
    uctx.register(
        PhysicalQuantity.COLLISION_COEFFICIENT,
        pinttr.UnitGenerator(lambda: uctx.get(PhysicalQuantity.LENGTH) ** -1),
    )
    uctx.register(
        PhysicalQuantity.INTENSITY,
        pinttr.UnitGenerator(
            lambda: unit_registry.watt
            / unit_registry.steradian
            / uctx.get(PhysicalQuantity.WAVELENGTH)
        ),
    )
    uctx.register(
        PhysicalQuantity.IRRADIANCE,
        pinttr.UnitGenerator(
            lambda: unit_registry.watt
            / uctx.get(PhysicalQuantity.LENGTH) ** 2
            / uctx.get(PhysicalQuantity.WAVELENGTH)
        ),
    )
    uctx.register(
        PhysicalQuantity.RADIANCE,
        pinttr.UnitGenerator(
            lambda: unit_registry.watt
            / uctx.get(PhysicalQuantity.LENGTH) ** 2
            / unit_registry.steradian
            / uctx.get(PhysicalQuantity.WAVELENGTH)
        ),
    )

    return uctx


#: Unit context used when interpreting config dictionaries.
#: Aliased in :mod:`eradiate`.
unit_context_config = _make_unit_context()

#: Unit context used when building kernel dictionaries.
#: Aliased in :mod:`eradiate`.
unit_context_kernel = _make_unit_context()


# -- Public functions ----------------------------------------------------------


def symbol(units: pint.Unit | str) -> str:
    """
    Normalize a string or Pint units to a symbol string.

    Parameters
    ----------
    units : :class:`pint.Unit` or str
        Value to convert to a symbol string.

    Returns
    -------
    str
        Symbol string (*e.g.* ``'m'`` for ``'metre'``, ``'W / m ** 2'`` for
        ``'W/m^2'``, etc.).
    """
    units = unit_registry.Unit(units)
    return format(units, "~")


def to_quantity(da: xarray.DataArray) -> pint.Quantity:
    """
    Converts a :class:`~xarray.DataArray` to a :class:`~pint.Quantity`.
    The array's ``attrs`` metadata mapping must contain a ``units`` field.

    Parameters
    ----------
    da : DataArray
        :class:`~xarray.DataArray` instance which will be converted.


    Returns
    -------
    quantity
        The corresponding Pint quantity.

    Raises
    ------
    ValueError
        If the array's metadata do not contain a ``units`` field.

    Notes
    -----
    This function can also be used on coordinate variables.
    """
    try:
        units = da.attrs["units"]
    except KeyError as e:
        raise ValueError("this DataArray has no 'units' metadata field") from e
    else:
        return unit_registry.Quantity(da.values, units)


def interpret_quantities(
    d: dict[str, t.Any],
    quantity_map: dict[str, str],
    uctx: pinttr.UnitContext,
    force: bool = False,
):
    """
    Advanced unit interpretation and wrapping for dictionaries. This function
    first calls :func:`pinttr.interpret_units` to interpret units attached to
    a given field. Then, it converts quantities and possibly applies default
    units to fields specified in ``quantity_map`` based on ``uctx``.

    Parameters
    ----------
    d : dict
        Dictionary to apply unit conversion, checking and defaults.

    quantity_map : dict[str, str]
        Dictionary mapping fields to quantity identifiers (see
        :class:`.PhysicalQuantity` for valid quantity IDs).

    uctx : :class:`pinttr.UnitContext`
        Unit context containing quantity and default units definitions.

    force : bool, default: False
        If ``True``, fields specified as quantities will be converted to target
        units; otherwise, only units compatibility will be checked.

    Returns
    -------
    dict
        Dictionary with units interpreted and checked, and default units
        applied to relevant fields.

    Raises
    ------
    :class:`pinttr.UnitsError`:
        If a field and its mapped quantity have incompatible units.

    See Also
    --------
    :class:`.PhysicalQuantity`
    """
    ureg = uctx.ureg

    # Interpret unit fields
    result = pinttr.interpret_units(d, ureg)

    # Convert to or apply default units based on the unit map
    if quantity_map is None:
        quantity_map = {}

    for key, quantity in quantity_map.items():
        value = result[key]
        if isinstance(value, pint.Quantity):
            units = uctx.get(quantity)
            if not units_compatible(value.units, units):
                raise UnitsError(value.units, units)

            if force:
                result[key] = value.to(units)
            else:
                result[key] = value
        else:
            result[key] = ureg.Quantity(result[key], uctx.get(quantity))

    return result



from ._access import load_dataset as load_dataset
from ._access import open_dataset as open_dataset
from ._access import open_datasets as open_datasets
from ._blind_directory import BlindDirectoryDataStore as BlindDirectoryDataStore
from ._blind_online import BlindOnlineDataStore as BlindOnlineDataStore
from ._core import DataStore as DataStore
from ._multi import MultiDataStore as MultiDataStore
from ._safe_directory import SafeDirectoryDataStore as SafeDirectoryDataStore
from ._safe_online import SafeOnlineDataStore as SafeOnlineDataStore
from ._store import data_store as data_store
from ._store import init_data_store as init_data_store



import numpy as np
from numpy.typing import ArrayLike


def square_to_uniform_disk_concentric(sample: ArrayLike) -> np.ndarray:
    """
    Low-distortion concentric square to disk mapping.

    Parameters
    ----------
    sample : array-like
        A (N, 2) array of sample values.

    Returns
    -------
    ndarray
        Sampled coordinates on the unit disk as a (N, 2) array.

    Notes
    -----
    The function tries to be flexible with arrays with (N, 1) and (N,) arrays
    and attempts reshaping them to (N/2, 2). This, in particular, means that
    the following call will produce the expected result:

    .. code:: python

       square_to_uniform_disk_concentric((0.5, 0.5))
    """
    # Matches Mitsuba implementation

    sample = np.atleast_1d(sample)
    if sample.ndim < 2:
        sample = sample.reshape((sample.size // 2, 2))
    if sample.ndim > 2 or sample.shape[1] != 2:
        raise ValueError(f"array must be of shape (N, 2), got {sample.shape}")

    x: ArrayLike = 2.0 * sample[..., 0] - 1.0
    y: ArrayLike = 2.0 * sample[..., 1] - 1.0

    is_zero = np.logical_and(x == 0.0, y == 0.0)
    quadrant_1_or_3 = np.abs(x) < np.abs(y)

    r = np.where(quadrant_1_or_3, y, x)
    rp = np.where(quadrant_1_or_3, x, y)

    phi = np.empty_like(r)
    phi[~is_zero] = 0.25 * np.pi * (rp[~is_zero] / r[~is_zero])
    phi[quadrant_1_or_3] = 0.5 * np.pi - phi[quadrant_1_or_3]
    phi[is_zero] = 0.0

    s, c = np.sin(phi), np.cos(phi)
    return np.vstack((r * c, r * s)).T


def square_to_uniform_hemisphere(sample: ArrayLike) -> np.ndarray:
    """
    Uniformly sample a vector on the unit hemisphere with respect to solid
    angles.

    Parameters
    ----------
    sample : array-like
        A (N, 2) array of sample values.

    Returns
    -------
    ndarray
        Sampled coordinates on the unit hemisphere as a (N, 3) array.

    Notes
    -----
    The function tries to be flexible with arrays with (N, 1) and (N,) arrays
    and attempts reshaping them to (N/2, 2). This, in particular, means that
    the following call will produce the expected result:

    .. code:: python

       square_to_uniform_hemisphere((0.5, 0.5))
    """
    # Matches Mitsuba implementation

    sample = np.atleast_1d(sample)
    if sample.ndim < 2:
        sample = sample.reshape((sample.size // 2, 2))
    if sample.ndim > 2 or sample.shape[1] != 2:
        raise ValueError(f"array must be of shape (N, 2), got {sample.shape}")

    p = square_to_uniform_disk_concentric(sample)
    z = 1.0 - np.multiply(p, p).sum(axis=1)
    p *= np.sqrt(z + 1.0).reshape((len(p), 1))
    return np.vstack((p[..., 0], p[..., 1], z)).T


def uniform_hemisphere_to_square(v: ArrayLike) -> np.ndarray:
    """
    Inverse of the mapping square_to_uniform_hemisphere.

    Parameters
    ----------
    v : array-like
        A (N, 3) array of vectors on the unit sphere.

    Returns
    -------
    ndarray
        Corresponding coordinates on the [0, 1]² square as a (N, 2) array.

    Notes
    -----
    The function tries to be flexible with arrays with (N, 1) and (N,) arrays
    and attempts reshaping them to (N/3, 3). This, in particular, means that
    the following call will produce the expected result:

    .. code:: python

       uniform_hemisphere_to_square((0, 0, 1))
    """
    # Matches Mitsuba implementation

    v = np.atleast_1d(v)
    if v.ndim < 2:
        v = v.reshape((v.size // 3, 3))
    if v.ndim > 2 or v.shape[1] != 3:
        raise ValueError(f"array must be of shape (N, 3), got {v.shape}")

    p = v[..., 0:2]
    return uniform_disk_to_square_concentric(
        p / np.sqrt(v[..., 2] + 1.0).reshape((len(p), 1))
    )


def uniform_disk_to_square_concentric(p: ArrayLike) -> np.ndarray:
    """
    Inverse of the mapping square_to_uniform_disk_concentric.

    Parameters
    ----------
    p : array-like
        A (N, 2) array of vectors on the unit disk.

    Returns
    -------
    ndarray
        Corresponding coordinates on the [0, 1]² square as a (N, 2) array.

    Notes
    -----
    The function tries to be flexible with arrays with (N, 1) and (N,) arrays
    and attempts reshaping them to (N/2, 2). This, in particular, means that
    the following call will produce the expected result:

    .. code:: python

       uniform_disk_to_square_concentric((0, 0))
    """
    # Matches Mitsuba implementation

    p = np.atleast_1d(p)
    if p.ndim < 2:
        p = p.reshape((p.size // 2, 2))
    if p.ndim > 2 or p.shape[1] != 2:
        raise ValueError(f"array must be of shape (N, 2), got {p.shape}")

    quadrant_0_or_2 = np.abs(p[..., 0]) > np.abs(p[..., 1])
    r_sign = np.where(quadrant_0_or_2, p[..., 0], p[..., 1])
    r = np.copysign(np.linalg.norm(p, axis=-1), r_sign)

    phi = np.arctan2(p[..., 1] * np.sign(r_sign), p[..., 0] * np.sign(r_sign))

    t = 4.0 / np.pi * phi
    t = np.where(quadrant_0_or_2, t, 2.0 - t) * r

    a = np.where(quadrant_0_or_2, r, t)
    b = np.where(quadrant_0_or_2, t, r)

    return np.vstack(((a + 1.0) * 0.5, (b + 1.0) * 0.5)).T



from __future__ import annotations

import logging
import typing as t
from abc import ABC, abstractmethod
from datetime import datetime

import attrs
import mitsuba as mi
import numpy as np
import pinttr
import xarray as xr
from hamilton.driver import Driver

import eradiate

from .. import converters, validators
from .. import pipelines as pl
from ..attrs import AUTO, documented, parse_docs
from ..contexts import KernelContext, MultiGenerator
from ..kernel import MitsubaObjectWrapper, mi_render, mi_traverse
from ..rng import SeedState
from ..scenes.atmosphere import AbstractHeterogeneousAtmosphere
from ..scenes.core import Scene, SceneElement, get_factory, traverse
from ..scenes.illumination import (
    AbstractDirectionalIllumination,
    ConstantIllumination,
    DirectionalIllumination,
    illumination_factory,
)
from ..scenes.integrators import Integrator, integrator_factory
from ..scenes.measure import (
    Measure,
    MultiDistantMeasure,
    measure_factory,
)
from ..spectral.ckd import BinSet, QuadSpec
from ..spectral.index import CKDSpectralIndex, MonoSpectralIndex, SpectralIndex
from ..spectral.mono import WavelengthSet
from ..units import unit_registry as ureg
from ..util.misc import deduplicate_sorted, onedict_value

logger = logging.getLogger(__name__)


def _convert_spectral_set(value):
    if value is AUTO:
        if eradiate.mode().is_ckd:
            return BinSet.default()
        elif eradiate.mode().is_mono:
            return WavelengthSet.default()
        else:
            raise NotImplementedError(f"unsupported mode: {eradiate.mode().id}")
    else:
        return value


@parse_docs
@attrs.define
class Experiment(ABC):
    """
    Abstract base class for all Eradiate experiments. An experiment consists of
    a high-level scene specification parametrized by natural user input, a
    processing and post-processing pipeline, and a result storage data
    structure.
    """

    # Internal Mitsuba scene. This member is not set by the end-user, but rather
    # by the Experiment itself during initialization.
    mi_scene: MitsubaObjectWrapper | None = attrs.field(
        default=None,
        repr=False,
        init=False,
    )

    measures: list[Measure] = documented(
        attrs.field(
            factory=lambda: [MultiDistantMeasure()],
            converter=lambda value: (
                [measure_factory.convert(x) for x in pinttr.util.always_iterable(value)]
                if not isinstance(value, dict)
                else [measure_factory.convert(value)]
            ),
            validator=attrs.validators.deep_iterable(
                member_validator=attrs.validators.instance_of(Measure)
            ),
        ),
        doc="List of measure specifications. The passed list may contain "
        "dictionaries, which will be interpreted by "
        ":data:`.measure_factory`. "
        "Optionally, a single :class:`.Measure` or dictionary specification "
        "may be passed and will automatically be wrapped into a list.",
        type="list of :class:`.Measure`",
        init_type="list of :class:`.Measure` or list of dict or "
        ":class:`.Measure` or dict",
        default=":class:`MultiDistantMeasure() <.MultiDistantMeasure>`",
    )

    integrator: Integrator = documented(
        attrs.field(
            default=AUTO,
            converter=converters.auto_or(integrator_factory.convert),
            validator=validators.auto_or(
                attrs.validators.instance_of(Integrator),
            ),
        ),
        doc="Monte Carlo integration algorithm specification. "
        "This parameter can be specified as a dictionary which will be "
        "interpreted by :data:`.integrator_factory`."
        "The integrator defaults to :data:`AUTO`, which will choose the appropriate "
        "integrator depending on the experiment's configuration. ",
        type=":class:`.Integrator` or AUTO",
        init_type=":class:`.Integrator` or dict or AUTO",
        default="AUTO",
    )

    _results: dict[str, xr.Dataset] = attrs.field(factory=dict, repr=False)

    @property
    def results(self) -> dict[str, xr.Dataset]:
        """
        Post-processed simulation results.

        Returns
        -------
        dict[str, Dataset]
            Dictionary mapping measure IDs to xarray datasets.
        """
        return self._results

    default_spectral_set: BinSet | WavelengthSet = documented(
        attrs.field(
            default=AUTO,
            validator=validators.auto_or(
                attrs.validators.instance_of((BinSet, WavelengthSet))
            ),
            converter=_convert_spectral_set,
            repr=False,
        ),
        doc="Default spectral set. This attribute is used to set the "
        "default value for :attr:`spectral_set`."
        "If the value is :data:`AUTO`, the default spectral set is selected "
        "based on the active mode. Otherwise, the value must be a "
        ":class:`.BinSet` or :class:`.WavelengthSet` instance.",
        type=":class:`.BinSet` or :class:`.WavelengthSet`",
        init_type=":class:`.BinSet` or :class:`.WavelengthSet` or :data:`AUTO`",
        default=":data:`AUTO`",
    )

    # Mapping of measure index and WavelengthSet or BinSet depending on active
    # mode. This attribute is set by the '_normalize_spectral()' method.
    _spectral_set = attrs.field(init=False, repr=False)

    @property
    def spectral_set(self) -> dict[int, WavelengthSet | BinSet]:
        """
        A dictionary mapping measure index to the associated spectral set.
        """
        return self._spectral_set

    quad_spec: QuadSpec = attrs.field(
        factory=QuadSpec.default,
        converter=QuadSpec.convert,
        validator=attrs.validators.instance_of(QuadSpec),
    )

    def __attrs_post_init__(self):
        self._normalize_spectral()

    def _normalize_spectral(self) -> None:
        """
        Setup spectral set based on active mode.
        """
        # default value for spectral set
        spectral_set = self.default_spectral_set

        # if the atmosphere provides a spectral set, use it
        atmosphere = getattr(self, "atmosphere", None)
        if atmosphere and isinstance(atmosphere, AbstractHeterogeneousAtmosphere):
            atmosphere_spectral_set = atmosphere.spectral_set(quad_spec=self.quad_spec)
            if atmosphere_spectral_set is not None:
                spectral_set = atmosphere_spectral_set

        # finally, the spectral set is filtered by the SRF
        self._spectral_set = {
            i: measure.srf.select_in(spectral_set)
            for i, measure in enumerate(self.measures)
        }

    def clear(self) -> None:
        """
        Clear previous experiment results and reset internal state.
        """
        self.results.clear()

        for measure in self.measures:
            measure.mi_results.clear()

    @abstractmethod
    def init(self) -> None:
        """
        Generate kernel dictionary and initialise Mitsuba scene.
        """
        pass

    @abstractmethod
    def process(
        self,
        spp: int = 0,
        seed_state: SeedState | None = None,
    ) -> None:
        """
        Run simulation and collect raw results.

        Parameters
        ----------
        spp : int, optional
            Sample count. If set to 0, the value set in the original scene
            definition takes precedence.

        seed_state : :class:`.SeedState`, optional
            Seed state used to generate seeds to initialize Mitsuba's RNG at
            every iteration of the parametric loop. If unset, Eradiate's
            :attr:`root seed state <.root_seed_state>` is used.
        """
        pass

    @abstractmethod
    def postprocess(self) -> None:
        """
        Post-process raw results and store them in :attr:`results`.
        """
        pass

    @abstractmethod
    def pipeline(self, measure: Measure | int) -> Driver:
        """
        Return the post-processing pipeline for a given measure.

        Parameters
        ----------
        measure : .Measure or int
            Measure for which the pipeline is generated.

        Returns
        -------
        hamilton.driver.Driver
        """
        pass

    @property
    @abstractmethod
    def context_init(self) -> KernelContext:
        """
        Return a single context used for scene initialization.
        """
        pass

    @property
    @abstractmethod
    def contexts(self) -> list[KernelContext]:
        """
        Return a list of contexts used for processing.
        """
        pass


def _extra_objects_converter(value: dict | None) -> dict:
    if not value:
        return {}

    result = {}

    for key, element_spec in value.items():
        if isinstance(element_spec, dict):
            element_spec = element_spec.copy()
            element_type = element_spec.pop("factory")
            factory = get_factory(element_type)
            result[key] = factory.convert(element_spec)

        else:
            result[key] = element_spec

    return result


@parse_docs
@attrs.define
class EarthObservationExperiment(Experiment, ABC):
    """
    Abstract based class for experiments illuminated by a distant directional
    emitter.
    """

    extra_objects: dict[str, SceneElement] = documented(
        attrs.field(
            default=None,
            converter=_extra_objects_converter,
            validator=attrs.validators.deep_mapping(
                key_validator=attrs.validators.instance_of(str),
                value_validator=attrs.validators.instance_of(SceneElement),
            ),
        ),
        doc="Dictionary of extra objects to be added to the scene. "
        "The keys of this dictionary are used to identify the objects "
        "in the kernel dictionary.",
        type="dict",
        init_type="dict or None",
        default="None",
    )

    illumination: AbstractDirectionalIllumination | ConstantIllumination = documented(
        attrs.field(
            factory=DirectionalIllumination,
            converter=illumination_factory.convert,
            validator=attrs.validators.instance_of(
                (AbstractDirectionalIllumination, ConstantIllumination)
            ),
        ),
        doc="Illumination specification. "
        "This parameter can be specified as a dictionary which will be "
        "interpreted by :data:`.illumination_factory`.",
        type=":class:`.AbstractDirectionalIllumination` or "
        ":class:`.ConstantIllumination`",
        init_type=":class:`.DirectionalIllumination` or "
        ":class:`.ConstantIllumination` or dict",
        default=":class:`DirectionalIllumination() <.DirectionalIllumination>`",
    )

    def _dataset_metadata(self, measure: Measure) -> dict[str, str]:
        """
        Generate additional metadata applied to dataset after post-processing.

        Parameters
        ----------
        measure : :class:`.Measure`
            Measure for which the metadata is created.

        Returns
        -------
        dict[str, str]
            Metadata to be attached to the produced dataset.
        """

        return {
            "convention": "CF-1.10",
            "source": f"eradiate, version {eradiate.__version__}",
            "history": f"{datetime.utcnow().replace(microsecond=0).isoformat()}"
            f" - data creation - {self.__class__.__name__}.postprocess()",
            "references": "",
        }

    def spectral_indices(self, measure_index: int) -> t.Generator[SpectralIndex]:
        """
        Generate spectral indices for a given measure.

        Parameters
        ----------
        measure_index : int
            Measure index for which spectral indices are generated.

        Yields
        ------
        :class:`.SpectralIndex`
            Spectral index.
        """
        if eradiate.mode().is_mono:
            generator = self.spectral_indices_mono
        elif eradiate.mode().is_ckd:
            generator = self.spectral_indices_ckd
        else:
            raise RuntimeError(f"unsupported mode '{eradiate.mode().id}'")

        yield from generator(measure_index)

    def spectral_indices_mono(
        self, measure_index: int
    ) -> t.Generator[MonoSpectralIndex]:
        yield from self.spectral_set[measure_index].spectral_indices()

    def spectral_indices_ckd(self, measure_index: int) -> t.Generator[CKDSpectralIndex]:
        yield from self.spectral_set[measure_index].spectral_indices()

    def _spectral_index_generator(self):
        return MultiGenerator(
            [self.spectral_indices(i) for i in range(len(self.measures))]
        )

    @property
    def context_init(self):
        return KernelContext(
            si=self._spectral_index_generator().__next__(), kwargs=self._context_kwargs
        )

    @property
    @abstractmethod
    def _context_kwargs(self) -> dict[str, t.Any]:
        pass

    @property
    def contexts(self) -> list[KernelContext]:
        # Inherit docstring

        # Collect contexts from all measures
        sis = []

        for measure_index, _ in enumerate(self.measures):
            _si = list(self.spectral_indices(measure_index))
            sis.extend(_si)

        # Sort and remove duplicates
        key = {
            MonoSpectralIndex: lambda si: si.w.m,
            CKDSpectralIndex: lambda si: (si.w.m, si.g),
        }[type(sis[0])]

        sis = deduplicate_sorted(
            sorted(sis, key=key), cmp=lambda x, y: key(x) == key(y)
        )
        kwargs = self._context_kwargs

        return [KernelContext(si, kwargs=kwargs) for si in sis]

    @property
    @abstractmethod
    def scene_objects(self) -> dict[str, SceneElement]:
        pass

    @property
    def scene(self) -> Scene:
        """
        Return a scene object used for kernel dictionary template and parameter
        table generation.
        """
        return Scene(objects={**self.scene_objects, **self.extra_objects})

    def init(self):
        # Inherit docstring

        logger.info("Initializing kernel scene")

        kdict_template, umap_template = traverse(self.scene)
        try:
            self.mi_scene = mi_traverse(
                mi.load_dict(kdict_template.render(ctx=self.context_init)),
                umap_template=umap_template,
            )
        except RuntimeError as e:
            raise RuntimeError(f"(while loading kernel scene dictionary){e}") from e

        # Remove unused elements from Mitsuba scene parameter table
        self.mi_scene.drop_parameters()

    def process(
        self,
        spp: int = 0,
        seed_state: SeedState | None = None,
    ) -> None:
        # Inherit docstring

        # Set up Mitsuba scene
        if self.mi_scene is None:
            self.init()

        # Run Mitsuba for each context
        logger.info("Launching simulation")

        mi_results = mi_render(
            self.mi_scene,
            self.contexts,
            seed_state=seed_state,
            spp=spp,
        )

        # Assign collected results to the appropriate measure
        sensor_to_measure: dict[str, Measure] = {
            measure.sensor_id: measure for measure in self.measures
        }

        for ctx_index, spectral_group_dict in mi_results.items():
            for sensor_id, mi_bitmap in spectral_group_dict.items():
                if self.integrator.moment:
                    split = mi_bitmap.split()
                    m2 = split[1][1]
                    m2_np = np.array(m2, copy=False)[:, :, [0]]
                    m2 = mi.Bitmap(m2_np, mi.Bitmap.PixelFormat.Y)

                    measure = sensor_to_measure[sensor_id]
                    measure.mi_results[ctx_index] = {
                        "bitmap": split[0][1],
                        "m2": m2,
                        "spp": spp if spp > 0 else measure.spp,
                    }

                else:
                    measure = sensor_to_measure[sensor_id]
                    measure.mi_results[ctx_index] = {
                        "bitmap": mi_bitmap,
                        "spp": spp if spp > 0 else measure.spp,
                    }

    def postprocess(self) -> None:
        # Inherit docstring
        logger.info("Post-processing results")
        measures = self.measures

        # Run pipelines
        for i, measure in enumerate(measures):
            drv: Driver = self.pipeline(measure)
            inputs = self._pipeline_inputs(i)
            outputs = pl.outputs(drv)
            result = drv.execute(final_vars=outputs, inputs=inputs)
            self.results[measure.id] = xr.Dataset({var: result[var] for var in outputs})

    def pipeline(self, measure: Measure | int) -> Driver:
        # Inherit docstring
        if isinstance(measure, int):
            measure = self.measures[measure]
        config = pl.config(measure, integrator=self.integrator)
        return eradiate.pipelines.driver(config)

    def _pipeline_inputs(self, i_measure: int):
        # This convenience function collects pipeline inputs for a specific measure

        measure = self.measures[i_measure]
        result = {
            "bitmaps": measure.mi_results,
            "spectral_set": self.spectral_set[i_measure],
            "illumination": self.illumination,
            "srf": measure.srf,
        }

        config = pl.config(measure)
        if config.get("add_viewing_angles", False):
            result["angles"] = measure.viewing_angles.m_as(ureg.deg)
        else:
            result["viewing_angles"] = None

        return result


# ------------------------------------------------------------------------------
#                              Experiment runner
# ------------------------------------------------------------------------------


def run(
    exp: Experiment,
    spp: int = 0,
    seed_state: SeedState | None = None,
) -> xr.Dataset | dict[str, xr.Dataset]:
    """
    Run an Eradiate experiment. This function performs kernel scene assembly,
    runs the computation and post-processes the raw results. The output consists
    of one or several xarray datasets.

    Parameters
    ----------
    exp : Experiment
        Reference to the experiment object which will be processed.

    spp : int, optional, default: 0
        Optional parameter to override the number of samples per pixel for all
        computed measures. If set to 0, the configured value for each measure
        takes precedence.

    seed_state : :class:`.SeedState`, optional
            Seed state used to generate seeds to initialize Mitsuba's RNG at
            every iteration of the parametric loop. If unset, Eradiate's
            :attr:`root seed state <.root_seed_state>` is used.

    Returns
    -------
    Dataset or dict[str, Dataset]
        If a single measure is defined, a single xarray dataset is returned.
        If several measures are defined, a dictionary mapping measure IDs to
        the corresponding result dataset is returned.
    """
    exp.process(spp=spp, seed_state=seed_state)
    exp.postprocess()
    return exp.results if len(exp.results) > 1 else onedict_value(exp.results)



from __future__ import annotations

import attrs
import mitsuba as mi
import numpy as np
import pint
import pinttr

from ._core import Measure
from ... import validators
from ...attrs import documented, parse_docs
from ...units import symbol
from ...units import unit_context_config as ucc
from ...units import unit_context_kernel as uck
from ...units import unit_registry as ureg


@parse_docs
@attrs.define(eq=False, slots=False)
class PerspectiveCameraMeasure(Measure):
    """
    Perspective camera scene element [``perspective``].

    This scene element is a thin wrapper around the ``perspective`` sensor
    kernel plugin. It positions a perspective camera based on a set of vectors,
    specifying the origin, viewing direction and 'up' direction of the camera.
    """

    # --------------------------------------------------------------------------
    #                           Fields and properties
    # --------------------------------------------------------------------------

    spp: int = documented(
        attrs.field(default=32, converter=int, validator=validators.is_positive),
        doc="Number of samples per pixel.",
        type="int",
        default="32",
    )

    _film_resolution: tuple[int, int] = documented(
        attrs.field(
            default=(32, 32),
            converter=tuple,
            validator=attrs.validators.deep_iterable(
                member_validator=attrs.validators.instance_of(int),
                iterable_validator=validators.has_len(2),
            ),
        ),
        doc="Film resolution as a (width, height) 2-tuple.",
        type="tuple of int",
        init_type="array-like",
        default="(32, 32)",
    )

    @property
    def film_resolution(self) -> tuple[int, int]:
        return self._film_resolution

    origin: pint.Quantity = documented(
        pinttr.field(
            factory=lambda: [1, 1, 1] * ureg.m,
            validator=[validators.has_len(3), pinttr.validators.has_compatible_units],
            units=ucc.deferred("length"),
        ),
        doc="A 3-vector specifying the position of the camera.\n"
        "\n"
        "Unit-enabled field (default: ucc['length']).",
        type="quantity",
        init_type="array-like",
        default="[1, 1, 1] m",
    )

    target: pint.Quantity = documented(
        pinttr.field(
            factory=lambda: [0, 0, 0] * ureg.m,
            validator=[validators.has_len(3), pinttr.validators.has_compatible_units],
            units=ucc.deferred("length"),
        ),
        doc="Point location targeted by the camera.\n"
        "\n"
        "Unit-enabled field (default: ucc['length']).",
        type="quantity",
        init_type="array-like",
        default="[0, 0, 0] m",
    )

    @target.validator
    @origin.validator
    def _target_origin_validator(self, attribute, value):
        if np.allclose(self.target, self.origin):
            raise ValueError(
                f"While initializing {attribute}: "
                f"Origin and target must not be equal, "
                f"got target = {self.target}, origin = {self.origin}"
            )

    up: np.ndarray = documented(
        attrs.field(
            factory=lambda: [0, 0, 1],
            converter=np.array,
            validator=validators.has_len(3),
        ),
        doc="A 3-vector specifying the up direction of the camera.\n"
        "This vector must be different from the camera's viewing direction,\n"
        "which is given by ``target - origin``.",
        type="array",
        default="[0, 0, 1]",
    )

    @up.validator
    def _up_validator(self, attribute, value):
        direction = self.target - self.origin
        if np.allclose(np.cross(direction, value), 0):
            raise ValueError(
                f"While initializing '{attribute.name}': "
                f"up direction must not be colinear with viewing direction, "
                f"got up = {self.up}, direction = {direction}"
            )

    far_clip: pint.Quantity = documented(
        pinttr.field(
            default=1e4 * ureg.km,
            units=ucc.deferred("length"),
        ),
        doc="Distance to the far clip plane.\n"
        "\n"
        "Unit-enabled field (default: ucc[length]).",
        type="quantity",
        init_type="quantity of float",
        default="10 000 km",
    )

    fov: pint.Quantity = documented(
        pinttr.field(default=50.0 * ureg.deg, units=ureg.deg),
        doc="Camera field of view.\n\nUnit-enabled field (default: degree).",
        type="quantity",
        init_type="quantity or float",
        default="50°",
    )

    # --------------------------------------------------------------------------
    #                       Kernel dictionary generation
    # --------------------------------------------------------------------------

    @property
    def kernel_type(self) -> str:
        # Inherit docstring
        return "perspective"

    @property
    def template(self) -> dict:
        # Inherit docstring
        result = super().template

        result["far_clip"] = self.far_clip.m_as(uck.get("length"))
        result["fov"] = self.fov.m_as(ureg.deg)

        target = self.target.m_as(uck.get("length"))
        origin = self.origin.m_as(uck.get("length"))
        result["to_world"] = mi.ScalarTransform4f.look_at(
            origin=origin, target=target, up=self.up
        )

        return result

    # --------------------------------------------------------------------------
    #                        Post-processing information
    # --------------------------------------------------------------------------

    @property
    def var(self) -> tuple[str, dict]:
        return "radiance", {
            "standard_name": "radiance",
            "long_name": "radiance",
            "units": symbol(uck.get("radiance")),
        }



from __future__ import annotations

import attrs

from ._core import Integrator
from ...attrs import documented, parse_docs


@parse_docs
@attrs.define(eq=False, slots=False)
class MonteCarloIntegrator(Integrator):
    """
    Base class for integrator elements wrapping kernel classes
    deriving from
    :class:`mitsuba.MonteCarloIntegrator`.

    .. warning:: This class should not be instantiated.
    """

    max_depth: int | None = documented(
        attrs.field(default=None, converter=attrs.converters.optional(int)),
        doc="Longest path depth in the generated measure data (where -1 "
        "corresponds to ∞). A value of 1 will display only visible emitters. 2 "
        "computes only direct illumination (no multiple scattering), etc. If "
        "unset, the kernel default value (-1) is used.",
        type="int or None",
        init_type="int, optional",
    )

    rr_depth: int | None = documented(
        attrs.field(default=None, converter=attrs.converters.optional(int)),
        doc="Minimum path depth after which the implementation starts applying "
        "the Russian roulette path termination criterion. If unset, the kernel "
        "default value (5) is used.",
        type="int or None",
        init_type="int, optional",
    )

    hide_emitters: bool | None = documented(
        attrs.field(default=None, converter=attrs.converters.optional(bool)),
        doc="Hide directly visible emitters. If unset, the kernel default "
        "value (``false``) is used.",
        type="bool or None",
        init_type="bool, optional",
    )

    @property
    def kernel_type(self) -> str:
        raise NotImplementedError

    @property
    def template(self) -> dict:
        result = {"type": self.kernel_type}

        if self.timeout is not None:
            result["timeout"] = self.timeout
        if self.max_depth is not None:
            result["max_depth"] = self.max_depth
        if self.rr_depth is not None:
            result["rr_depth"] = self.rr_depth
        if self.hide_emitters is not None:
            result["hide_emitters"] = self.hide_emitters

        if self.moment:
            result = { "type":"moment", "nested":result }
        
        return result


@parse_docs
@attrs.define(eq=False, slots=False)
class PathIntegrator(MonteCarloIntegrator):
    """
    A thin interface to the path tracer kernel plugin.

    This integrator samples paths using random walks starting from the sensor.
    It supports multiple scattering and does not account for volume
    interactions.
    """

    @property
    def kernel_type(self) -> str:
        return "path"


@parse_docs
@attrs.define(eq=False, slots=False)
class VolPathIntegrator(MonteCarloIntegrator):
    """
    A thin interface to the volumetric path tracer kernel plugin.

    This integrator samples paths using random walks starting from the sensor.
    It supports multiple scattering and accounts for volume interactions.
    """

    @property
    def kernel_type(self) -> str:
        return "volpath"


@parse_docs
@attrs.define(eq=False, slots=False)
class VolPathMISIntegrator(MonteCarloIntegrator):
    """
    A thin interface to the volumetric path tracer (with spectral multiple
    importance sampling) kernel plugin
    :cite:`Miller2019NullscatteringPathIntegral`.
    """

    use_spectral_mis = attrs.field(
        default=None, converter=attrs.converters.optional(bool)
    )

    @property
    def kernel_type(self) -> str:
        return "volpathmis"

    @property
    def template(self) -> dict:
        result = super().template
        if self.use_spectral_mis is not None:
            result["use_spectral_mis"] = self.use_spectral_mis

        return result

@parse_docs
@attrs.define(eq=False, slots=False)
class PiecewiseVolPathIntegrator(MonteCarloIntegrator):
    """
    A thin interface to the piecewise volumetric path tracer kernel plugin.

    This integrator samples paths using random walks starting from the sensor.
    It supports multiple scattering and accounts for 1D volume interactions.
    """

    @property
    def kernel_type(self) -> str:
        return "piecewise_volpath"



from __future__ import annotations

import typing as t
from abc import ABC, abstractmethod

import attrs
import numpy as np
import pint
import pinttr

from ._distant import AbstractDistantMeasure
from ... import converters, frame
from ...attrs import documented, parse_docs
from ...config import settings
from ...units import symbol
from ...units import unit_context_config as ucc
from ...units import unit_context_kernel as uck
from ...units import unit_registry as ureg
from ...util.misc import flatten

# ------------------------------------------------------------------------------
#                               Layout framework
# ------------------------------------------------------------------------------


@parse_docs
@attrs.define
class Layout(ABC):
    """
    Abstract base class for all viewing direction layouts.
    """

    azimuth_convention: frame.AzimuthConvention = documented(
        attrs.field(
            default=None,
            converter=lambda x: (
                settings.azimuth_convention
                if x is None
                else frame.AzimuthConvention.convert(x)
            ),
            validator=attrs.validators.instance_of(frame.AzimuthConvention),
            kw_only=True,
        ),
        doc="Azimuth convention used by this layout. If ``None``, the global "
        "default configuration is used (see :ref:`sec-user_guide-config`).",
        type=".AzimuthConvention",
        init_type=".AzimuthConvention or str, optional",
        default="None",
    )

    @staticmethod
    def convert(value: t.Any) -> t.Any:
        """
        Attempt to instantiate a :class:`Layout` concrete class from an object.

        This conversion protocol accepts:

        * a dictionary of the form ``{"type": type_name, **kwargs}``;
        * a (N, 2)-array or a (2,)-array;
        * a (N, 3)-array or a (3,)-array.

        Other values pass through the converter.

        Dictionaries have their parameters forwarded to the type selected by the
        ``type`` parameter.
        A (N, 2) or (2,)-array is passed to an :class:`.AngleLayout`.
        A (N, 3) or (3,)-array is passed to a :class:`.DirectionLayout`.

        .. list-table::
           :header-rows: 1

           * - Type key
             - Class
           * - angles
             - :class:`.AngleLayout`
           * - aring
             - :class:`.AzimuthRingLayout`
           * - directions
             - :class:`.DirectionLayout`
           * - grid
             - :class:`.GridLayout`
           * - hplane
             - :class:`.HemispherePlaneLayout`
        """
        if isinstance(value, Layout):
            return value

        if isinstance(value, dict):
            d = pinttr.interpret_units(value, ureg=ureg)
            type_key = d.pop("type")
            cls = {
                "angles": AngleLayout,
                "aring": AzimuthRingLayout,
                "directions": DirectionLayout,
                "grid": GridLayout,
                "hplane": HemispherePlaneLayout,
            }[type_key]
            return cls(**d)

        if np.ndim(value) == 2:
            if np.shape(value)[1] == 2:
                return AngleLayout(angles=value)
            if np.shape(value)[1] == 3:
                return DirectionLayout(directions=value)

        if np.ndim(value) == 1:
            if np.shape(value) == (2,):
                return AngleLayout(angles=value)
            if np.shape(value) == (3,):
                return DirectionLayout(directions=value)

        return value

    @property
    def n_directions(self) -> int:
        """
        int: Number of viewing directions defined by this layout.
        """
        return len(self.angles)

    @property
    @abstractmethod
    def angles(self) -> pint.Quantity:
        """
        quantity: A sequence of viewing angles, corresponding to the direction
            sequence produced by :attr:`directions`, as a (N, 2) array.
            The last dimension is ordered as (zenith, azimuth).
        """
        pass

    @property
    def directions(self) -> np.narray:
        """
        ndarray: A sequence of viewing directions, pointing *outwards* the
            observed target, as a (N, 3) array.
        """
        # Default implementation computes directions from angles
        return frame.angles_to_direction(
            self.angles, azimuth_convention=self.azimuth_convention
        )


def _angles_converter(value):
    value = pinttr.util.ensure_units(value, ucc.deferred("angle"))
    angle_units = value.u
    magnitude = np.reshape(value.m_as(ureg.deg), (-1, 2))
    zeniths = magnitude[:, 0]
    azimuths = magnitude[:, 1]
    return (np.stack((zeniths, azimuths % 360), axis=1) * ureg.deg).to(angle_units)


@parse_docs
@attrs.define
class AngleLayout(Layout):
    """
    A viewing direction layout directly defined by explicit (zenith, azimuth)
    pairs.
    """

    _angles: pint.Quantity = documented(
        pinttr.ib(
            converter=_angles_converter,
            units=ucc.deferred("angle"),
        ),
        doc="A sequence of viewing angles, corresponding to the direction "
        "sequence produced by :attr:`directions`, as a (N, 2) array. "
        "The last dimension is ordered as (zenith, azimuth). Zenith values "
        "must be between 0 and 180°. **Required, no default**.\n"
        "\n"
        "Unit-enabled field (default: ucc['angle']).",
        type="quantity",
        init_type="array-like",
    )

    @_angles.validator
    def _angles_validator(self, attribute, value):
        zeniths = value[:, 0].m_as(ureg.deg)
        if np.any((zeniths < 0) | (zeniths > 180)):
            raise ValueError(
                f"while validating {attribute.name}: zenith values must be in "
                "[0°, 180°]"
            )

    @property
    def angles(self) -> pint.Quantity:
        # Inherit docstring
        return self._angles


@parse_docs
@attrs.define
class AzimuthRingLayout(Layout):
    """
    A viewing direction layout defined by a single zenith and a vector of
    explicit azimuth values.
    """

    zenith: pint.Quantity = documented(
        pinttr.field(
            converter=attrs.converters.pipe(
                pinttr.converters.to_units(ucc.deferred("angle")),
                lambda x: converters.on_quantity(float)(x),
            ),
            units=ucc.deferred("angle"),
        ),
        doc="A single zenith value. Must be in [0°, 180°]. **Required, no default**.\n"
        "\n"
        "Unit-enabled field (default: ucc['angle']).",
        type="quantity",
        init_type="float or quantity",
    )

    @zenith.validator
    def _zenith_validator(self, attribute, value):
        zenith = value.m_as(ureg.deg)
        if zenith < 0 or zenith > 180:
            raise ValueError(
                f"while validating {attribute.name}: zenith value must be in [0°, 180°]"
            )

    azimuths: pint.Quantity = documented(
        pinttr.field(
            converter=attrs.converters.pipe(
                pinttr.converters.to_units(ucc.deferred("angle")),
                lambda x: np.reshape(x, (-1,)),
                lambda x: x % (360.0 * ureg.deg),
            ),
            units=ucc.deferred("angle"),
        ),
        doc="A vector of azimuth values. **Required, no default**.\n"
        "\n"
        "Unit-enabled field (default: ucc['angle']).",
        type="quantity",
        init_type="array-like",
    )

    @property
    def angles(self) -> pint.Quantity:
        # Inherit docstring
        # Basic unit conversion and broadcasting
        angle_units = ucc.get("angle")
        azimuths = np.reshape(self.azimuths.m_as(angle_units), (-1, 1))
        zeniths = np.full_like(azimuths, self.zenith.m_as(angle_units))

        # Assemble angles
        return np.hstack((zeniths, azimuths)) * angle_units


@parse_docs
@attrs.define
class DirectionLayout(Layout):
    """
    A viewing direction layout directly defined by explicit (zenith, azimuth)
    pairs.
    """

    _directions: np.ndarray = documented(
        attrs.field(converter=lambda x: np.reshape(x, (-1, 3))),
        doc="A sequence of 3-vectors specifying distant sensing directions. "
        "Note that directions point outward the target. **Required, no default**.",
        type="ndarray",
        init_type="array-like",
    )

    @property
    def angles(self) -> pint.Quantity:
        # Inherit docstring

        return frame.direction_to_angles(
            self.directions,
            azimuth_convention=self.azimuth_convention,
            normalize=True,
        ).to(ucc.get("angle"))  # Convert to default angle units

    @property
    def n_directions(self) -> int:
        # Inherit docstring
        return len(self._directions)

    @property
    def directions(self) -> np.narray:
        # Inherit docstring
        return self._directions


@parse_docs
@attrs.define
class HemispherePlaneLayout(Layout):
    """
    A viewing direction layout defined by a single azimuth and a vector of
    zenith values. Negative zenith values are mapped to (azimuth + 180°).
    """

    zeniths: pint.Quantity = documented(
        pinttr.field(
            converter=attrs.converters.pipe(
                pinttr.converters.to_units(ucc.deferred("angle")),
                lambda x: np.reshape(x, (-1,)),
            ),
            units=ucc.deferred("angle"),
        ),
        doc="A vector of zenith values. **Required, no default**.\n"
        "\n"
        "Unit-enabled field (default: ucc['angle']).",
        type="quantity",
        init_type="array-like",
    )

    azimuth: pint.Quantity = documented(
        pinttr.field(
            converter=attrs.converters.pipe(
                pinttr.converters.to_units(ucc.deferred("angle")),
                lambda x: x % (360 * ureg.deg),
            ),
            units=ucc.deferred("angle"),
        ),
        doc="A single zenith value. **Required, no default**.",
        type="quantity",
        init_type="float or quantity",
    )

    @property
    def angles(self) -> pint.Quantity:
        # Inherit docstring
        # Basic unit conversion and broadcasting
        angle_units = ucc.get("angle")
        zeniths = np.reshape(self.zeniths.m_as(angle_units), (-1, 1))
        azimuths = np.full_like(zeniths, self.azimuth.m_as(angle_units))

        # Assemble angles
        return np.hstack((zeniths, azimuths)) * angle_units


@parse_docs
@attrs.define
class GridLayout(Layout):
    """
    A viewing direction layout defined as the Cartesian product of an azimuth
    and zenith vectors.
    """

    zeniths: pint.Quantity = documented(
        pinttr.field(
            converter=attrs.converters.pipe(
                pinttr.converters.to_units(ucc.deferred("angle")),
                lambda x: np.reshape(x, (-1,)),
            ),
            units=ucc.deferred("angle"),
        ),
        doc="A vector of zenith values. **Required, no default**.\n"
        "\n"
        "Unit-enabled field (default: ucc['angle']).",
        type="quantity",
        init_type="array-like",
    )

    @zeniths.validator
    def _zeniths_validator(self, attribute, value):
        zeniths = value.m_as(ureg.deg)
        if np.any((zeniths < 0) | (zeniths > 180)):
            raise ValueError(
                f"while validating {attribute.name}: zenith values must be in "
                "[0°, 180°]"
            )

    azimuths: pint.Quantity = documented(
        pinttr.field(
            converter=attrs.converters.pipe(
                pinttr.converters.to_units(ucc.deferred("angle")),
                lambda x: np.reshape(x, (-1,)),
            ),
            units=ucc.deferred("angle"),
        ),
        doc="A vector of azimuth values. **Required, no default**.\n"
        "\n"
        "Unit-enabled field (default: ucc['angle']).",
        type="quantity",
        init_type="array-like",
    )

    @property
    def angles(self) -> pint.Quantity:
        # Inherit docstring
        #  Basic unit conversion and broadcasting
        angle_units = ucc.get("angle")
        zeniths = self.zeniths.m_as(angle_units)
        azimuths = self.azimuths.m_as(angle_units)

        # Assemble angles
        # This effectively produces the Cartesian product of the zeniths and
        # azimuths arrays (see https://stackoverflow.com/a/11146645/3645374)
        return np.dstack(np.meshgrid(zeniths, azimuths)).reshape(-1, 2) * angle_units


# ------------------------------------------------------------------------------
#                       MultiDistantMeasure implementation
# ------------------------------------------------------------------------------


def _extract_kwargs(kwargs: dict, keys: list[str]) -> dict:
    # Helper function to collect common layout keyword arguments
    # (mutates the param dictionary)
    # Used in MultiDistantMeasure constructors
    return {key: kwargs.pop(key) for key in keys if key in kwargs}


@parse_docs
@attrs.define(eq=False, slots=False)
class MultiDistantMeasure(AbstractDistantMeasure):
    """
    Multi-distant radiance measure scene element [``mdistant``, ``multi_distant``].

    This scene element creates a measure consisting of an array of
    radiancemeters positioned at an infinite distance from the scene. In
    practice, it can be used to compute the radiance leaving a scene at the
    top of the atmosphere (or canopy if there is no atmosphere). Coupled to
    appropriate post-processing operations, scene reflectance can be derived
    from the radiance values it produces.

    .. admonition:: Class method constructors

       .. autosummary::

          aring
          grid
          hplane
          from_angles
          from_directions

    Notes
    -----
    * Setting the ``target`` parameter is required to get meaningful results.
      Experiment classes should take care of setting it appropriately.
    """

    # --------------------------------------------------------------------------
    #                           Fields and properties
    # --------------------------------------------------------------------------

    direction_layout: Layout = documented(
        attrs.field(
            kw_only=True,
            factory=lambda: DirectionLayout(directions=[0, 0, 1]),
            converter=Layout.convert,
            validator=attrs.validators.instance_of(Layout),
        ),
        doc="A viewing direction layout. Specification through a dictionary or "
        "arrays, as documented by :meth:`Layout.convert`, is also possible. "
        "The constructor methods provide a convenient interface to configure "
        "this parameter automatically.",
        type=".Layout",
        init_type="dict or array-like or .Layout, optional",
        default="DirectionLayout(directions=[0, 0, 1])",
    )

    @property
    def viewing_angles(self) -> pint.Quantity:
        """
        quantity: Viewing angles computed from stored `directions` as a
            (N, 1, 2) array, where N is the number of directions. The last
            dimension is ordered as (zenith, azimuth).
        """
        # Note: The middle dimension in (N, 1, 2) is the film height
        return self.direction_layout.angles.reshape(-1, 1, 2)

    @property
    def film_resolution(self) -> tuple[int, int]:
        # Inherit docstring
        return (self.direction_layout.n_directions, 1)

    # --------------------------------------------------------------------------
    #                         Additional constructors
    # --------------------------------------------------------------------------

    @classmethod
    def hplane(
        cls,
        zeniths: np.typing.ArrayLike,
        azimuth: float | pint.Quantity,
        **kwargs,
    ) -> MultiDistantMeasure:
        """
        Construct using a hemisphere plane cut viewing direction layout.

        Parameters
        ----------
        zeniths : array-like
            List of zenith values. Negative values are mapped to the
            `azimuth + 180°` half-plane. Unitless values are converted to
            ``ucc['angle']``.

        azimuth : float or quantity
            Hemisphere plane cut azimuth value. Unitless values are converted to
            ``ucc['angle']``.

        azimuth_convention : .AzimuthConvention or str, optional
            The azimuth convention applying to the viewing direction layout.
            If unset, the global default convention is used.

        **kwargs
            Remaining keyword arguments are forwarded to the
            :class:`.MultiDistantMeasure` constructor.

        Returns
        -------
        MultiDistantMeasure
        """
        layout = HemispherePlaneLayout(
            zeniths=zeniths,
            azimuth=azimuth,
            **_extract_kwargs(kwargs, ["azimuth_convention"]),
        )
        return cls(direction_layout=layout, **kwargs)

    @classmethod
    def aring(
        cls,
        zenith: float | pint.Quantity,
        azimuths: np.typing.ArrayLike,
        **kwargs,
    ) -> MultiDistantMeasure:
        """
        Construct using an azimuth ring viewing direction layout.

        Parameters
        ----------
        zenith : float or quantity
            Azimuth ring zenith value. Unitless values are converted to
            ``ucc['angle']``.

        azimuths : array-like
            List of azimuth values. Unitless values are converted to
            ``ucc['angle']``.

        azimuth_convention : .AzimuthConvention or str, optional
            The azimuth convention applying to the viewing direction layout.
            If unset, the global default convention is used.

        **kwargs
            Remaining keyword arguments are forwarded to the
            :class:`.MultiDistantMeasure` constructor.

        Returns
        -------
        MultiDistantMeasure
        """
        layout = AzimuthRingLayout(
            zenith=zenith,
            azimuths=azimuths,
            **_extract_kwargs(kwargs, ["azimuth_convention"]),
        )
        return cls(direction_layout=layout, **kwargs)

    @classmethod
    def grid(
        cls, zeniths: np.typing.ArrayLike, azimuths: np.typing.ArrayLike, **kwargs
    ) -> MultiDistantMeasure:
        """
        Construct using a gridded viewing direction layout, defined as the
        Cartesian product of zenith and azimuth arrays.

        Parameters
        ----------
        azimuths : array-like
            List of azimuth values.

        zeniths : array-like
            List of zenith values.

        azimuth_convention : .AzimuthConvention or str, optional
            The azimuth convention applying to the viewing direction layout.
            If unset, the global default convention is used.

        **kwargs
            Remaining keyword arguments are forwarded to the
            :class:`.MultiDistantMeasure` constructor.

        Returns
        -------
        MultiDistantMeasure
        """
        layout = GridLayout(
            zeniths=zeniths,
            azimuths=azimuths,
            **_extract_kwargs(kwargs, ["azimuth_convention"]),
        )
        return cls(direction_layout=layout, **kwargs)

    @classmethod
    def from_angles(cls, angles: np.typing.ArrayLike, **kwargs) -> MultiDistantMeasure:
        """
        Construct using a direction layout defined by explicit (zenith, azimuth)
        pairs.

        Parameters
        ----------
        angles : array-like
            A sequence of (zenith, azimuth), interpreted as (N, 2)-shaped array.

        azimuth_convention : .AzimuthConvention or str, optional
            The azimuth convention applying to the viewing direction layout.
            If unset, the global default convention is used.

        **kwargs
            Remaining keyword arguments are forwarded to the
            :class:`.MultiDistantMeasure` constructor.

        Returns
        -------
        MultiDistantMeasure
        """
        layout = AngleLayout(
            angles=angles,
            **_extract_kwargs(kwargs, ["azimuth_convention"]),
        )
        return cls(direction_layout=layout, **kwargs)

    @classmethod
    def from_directions(
        cls, directions: np.typing.ArrayLike, **kwargs
    ) -> MultiDistantMeasure:
        """
        Construct using a direction layout defined by explicit direction
        vectors.

        Parameters
        ----------
        directions : array-like
            A sequence of direction vectors, interpreted as (N, 3)-shaped array.

        azimuth_convention : .AzimuthConvention or str, optional
            The azimuth convention applying to the viewing direction layout.
            If unset, the global default convention is used.

        **kwargs
            Remaining keyword arguments are forwarded to the
            :class:`.MultiDistantMeasure` constructor.

        Returns
        -------
        MultiDistantMeasure

        Warnings
        --------
        Viewing directions are defined pointing *outwards* the target location.
        """
        layout = DirectionLayout(
            directions=directions,
            **_extract_kwargs(kwargs, ["azimuth_convention"]),
        )
        return cls(direction_layout=layout, **kwargs)

    # --------------------------------------------------------------------------
    #                       Kernel dictionary generation
    # --------------------------------------------------------------------------

    @property
    def kernel_type(self) -> str:
        return "mdistant"

    @property
    def template(self) -> dict:
        result = super().template
        result["directions"] = ",".join(
            map(str, -self.direction_layout.directions.ravel(order="C"))
        )

        if self.target is not None:
            result.update(flatten({"target": self.target.kernel_item()}))

        if self.ray_offset is not None:
            result["ray_offset"] = self.ray_offset.m_as(uck.get("length"))

        return result

    # --------------------------------------------------------------------------
    #                        Post-processing information
    # --------------------------------------------------------------------------

    @property
    def var(self) -> tuple[str, dict]:
        # Inherit docstring
        return "radiance", {
            "standard_name": "radiance",
            "long_name": "radiance",
            "units": symbol(uck.get("radiance")),
        }



import numpy as np

from eradiate.data import data_store
from eradiate.experiments import CanopyExperiment
from eradiate.units import unit_registry as ureg


def fetch_het01_brfpp():
    r"""
    Fetches data necessary for test case het01_brfpp. 
    Note that this function can be used to prefetch data 
    before running a test case.

    Returns
    -------
    dict
        dictionary containing the various data stores.
    """

    data = {}
    data["leaf_spec_path"] = data_store.fetch(
        "tests/regression_test_specifications/het01/het01_UNI_sphere.def"
    )
    data["leaf_pos_path"] = data_store.fetch(
        "tests/regression_test_specifications/het01/het01_UNI_instances.def"
    )

    return data

def create_het01_brfpp(prefetched_data = None):
    r"""
    Floating spheres (HET01)
    ========================

    Rationale
    ---------

    This test case implements a basic canopy scene:

    * Surface with lambertian reflectance
    * No atmosphere
    * Three dimensional canopy

    Parameters

    * Surface: Square surface with labmertian BSDF with :math:`r = 0.159`
    * Canopy: Floating spheres made up of disks with bilambertian bsdf model
      Leaf reflectance is 0.4957, transmittance is 0.4409.
      Disk and sphere positioning follow the HET01 scenario of the RAMI-3 benchmark
    * Illumination: Directional illumination with a zenith angle :math:`\theta = 20°`
    * Sensor: Distant reflectance measure, covering a plane, (76 angular points)

    """
    if prefetched_data is None :
        prefetched_data = fetch_het01_brfpp()

    leaf_spec_path = prefetched_data["leaf_spec_path"]
    leaf_pos_path = prefetched_data["leaf_pos_path"]

    return CanopyExperiment(
        canopy={
            "type": "discrete_canopy",
            "instanced_canopy_elements": [
                {
                    "construct": "from_file",
                    "filename": leaf_pos_path,
                    "canopy_element": {
                        "type": "leaf_cloud",
                        "construct": "from_file",
                        "filename": leaf_spec_path,
                        "leaf_reflectance": 0.4957,
                        "leaf_transmittance": 0.4409,
                        "id": "spherical_leaf_cloud",
                    },
                },
            ],
            "size": [100, 100, 30] * ureg.m,
        },
        surface={"type": "lambertian", "reflectance": 0.159},
        padding=20,
        measures=[
            {
                "type": "mdistant",
                "construct": "hplane",
                "spp": 10000,
                "azimuth": 180 * ureg.deg,
                "zeniths": np.arange(-75, 75.01, 2) * ureg.deg,
                "target": {
                    "type": "rectangle",
                    "xmin": -50 * ureg.m,
                    "xmax": 50 * ureg.m,
                    "ymin": -50 * ureg.m,
                    "ymax": 50 * ureg.m,
                    "z": 30 * ureg.m,
                },
            }
        ],
        illumination={
            "type": "directional",
            "zenith": 20 * ureg.deg,
            "azimuth": 0.0 * ureg.deg,
            "irradiance": 20.0,
        },
        integrator={"type": "path", "max_depth": -1},
    )

def fetch_het04a1_brfpp():
    r"""
    Fetches data necessary for test case het04_brfpp. 
    Note that this function can be used to prefetch data 
    before running a test case.

    Returns
    -------
    dict
        dictionary containing the various data stores.
    """

    data = {}

    data["leaf_spec_path_sph"] = data_store.fetch(
        "tests/regression_test_specifications/het04/het04_sph.def"
    )
    data["leaf_pos_path_sph"] = data_store.fetch(
        "tests/regression_test_specifications/het04/het04_sph_positions.def"
    )
    data["leaf_spec_path_cyl"] = data_store.fetch(
        "tests/regression_test_specifications/het04/het04_cyl.def"
    )
    data["leaf_pos_path_cyl"] = data_store.fetch(
        "tests/regression_test_specifications/het04/het04_cyl_positions.def"
    )
    
    return data

def create_het04a1_brfpp(prefetched_data = None):
    r"""
    Real zoom in (HET04a1)
    ======================

    This is a regression test, which compares the simulation results of the
    current branch to an older reference version.

    Rationale
    ---------

    This test case implements a basic canopy scene:

    * Surface with lambertian reflectance
    * No atmosphere
    * Three dimensional canopy

    Parameters

    * Surface: Square surface with labmertian BSDF with :math:`r = 0.15`
    * Canopy:

      * Floating spheres made up of disks with bilambertian bsdf model.
        Leaf reflectance is 0.49, transmittance is 0.41.
      * Floating cylinders made up of disks with bilambertian bsdf model.
        Leaf reflectance is 0.45, transmittance is 0.43.

      Disk, sphere and cylinder positioning follow the HET04 scenario of the
      RAMI-3 benchmark.
    * Illumination: Directional illumination with a zenith angle :math:`\theta = 20°`
    * Sensor: Distant reflectance measure, covering a plane, (76 angular points). 
      This test implements the variant a1 of the HET04
      scenario, in which the entire scene is targeted by the sensor.

    """

    if prefetched_data is None :
        prefetched_data = fetch_het04a1_brfpp()

    leaf_spec_path_sph = prefetched_data["leaf_spec_path_sph"]
    leaf_pos_path_sph = prefetched_data["leaf_pos_path_sph"]
    leaf_spec_path_cyl = prefetched_data["leaf_spec_path_cyl"]
    leaf_pos_path_cyl = prefetched_data["leaf_pos_path_cyl"]

    return CanopyExperiment(
        canopy={
            "type": "discrete_canopy",
            "instanced_canopy_elements": [
                {
                    "construct": "from_file",
                    "filename": leaf_pos_path_sph,
                    "canopy_element": {
                        "type": "leaf_cloud",
                        "construct": "from_file",
                        "filename": leaf_spec_path_sph,
                        "leaf_reflectance": 0.49,
                        "leaf_transmittance": 0.41,
                        "id": "spherical_leaf_cloud",
                    },
                },
                {
                    "construct": "from_file",
                    "filename": leaf_pos_path_cyl,
                    "canopy_element": {
                        "type": "leaf_cloud",
                        "construct": "from_file",
                        "filename": leaf_spec_path_cyl,
                        "leaf_reflectance": 0.45,
                        "leaf_transmittance": 0.3,
                        "id": "cylindrical_leaf_cloud",
                    },
                },
            ],
            "size": [270, 270, 15] * ureg.m,
        },
        surface={"type": "lambertian", "reflectance": 0.15},
        padding=5,
        measures=[
            {
                "type": "mdistant",
                "construct": "hplane",
                "spp": 20000,
                "azimuth": 180 * ureg.deg,
                "zeniths": np.arange(-75, 75.01, 2) * ureg.deg,
                "target": {
                    "type": "rectangle",
                    "xmin": -135 * ureg.m,
                    "xmax": 135 * ureg.m,
                    "ymin": -135 * ureg.m,
                    "ymax": 135 * ureg.m,
                    "z": 15 * ureg.m,
                },
            }
        ],
        illumination={
            "type": "directional",
            "zenith": 20 * ureg.deg,
            "azimuth": 0.0 * ureg.deg,
            "irradiance": 20.0,
        },
        integrator={"type": "path", "max_depth": -1},
    )

def fetch_het06_brfpp():
    r"""
    Fetches data necessary for test case het04_brfpp. 
    Note that this function can be used to prefetch data 
    before running a test case.

    Returns
    -------
    dict
        dictionary containing the various data stores.
    """

    data = {}

    data["tree_pos_path"] = data_store.fetch(
        "tests/regression_test_specifications/het06/het06_scene.def"
    )

    return data

def create_het06_brfpp(prefetched_data = None):
    r"""
    Coniferous forest (HET06)
    =========================

    This is a regression test, which compares the simulation results of the
    current branch to an older reference version.

    Rationale
    ---------

    This test case implements a basic canopy scene:

    * Surface with lambertian reflectance
    * No atmosphere
    * Three dimensional canopy

    Parameters

    * Surface: Square surface with labmertian BSDF with :math:`r = 0.86`
    * Canopy: Coniferous trees made up of a conical leaf cloud and a cylindrical
      trunk.

      * Leaf reflectance: 0.08
      * Leaf transmittance: 0.03
      * Trunk reflectance: 0.14

      Disk and tree positioning follow the HET06 scenario of the RAMI-3 benchmark
    * Illumination: Directional illumination with a zenith angle :math:`\theta = 20°`
    * Sensor: Distant reflectance measure, covering a plane, (76 angular points).

    """

    if prefetched_data is None :
        prefetched_data = fetch_het06_brfpp()

    tree_pos_path = prefetched_data["tree_pos_path"]

    return CanopyExperiment(
        canopy={
            "type": "discrete_canopy",
            "instanced_canopy_elements": [
                {
                    "construct": "from_file",
                    "filename": tree_pos_path,
                    "canopy_element": {
                        "type": "abstract_tree",
                        "trunk_height": 1.5 * ureg.m,
                        "trunk_radius": 0.15 * ureg.m,
                        "trunk_reflectance": 0.14,
                        "leaf_cloud_extra_offset": [0.0, 0.0, 0.0] * ureg.m,
                        "id": "conifer_tree",
                        "leaf_cloud": {
                            "construct": "cone",
                            "l_vertical": 12.0 * ureg.m,
                            "leaf_radius": 0.05,
                            "radius": 1.8 * ureg.m,
                            "n_leaves": 6480,
                            "leaf_reflectance": 0.08,
                            "leaf_transmittance": 0.03,
                            "id": "conical_leaf_cloud",
                        },
                    },
                }
            ],
            "size": [500, 500, 13.5] * ureg.m,
        },
        surface={"type": "lambertian", "reflectance": 0.86},
        padding=0,
        measures=[
            {
                "type": "mdistant",
                "construct": "hplane",
                "spp": 10000,
                "azimuth": 180 * ureg.deg,
                "zeniths": np.arange(-75, 75.01, 2) * ureg.deg,
                "target": {
                    "type": "rectangle",
                    "xmin": -250 * ureg.m,
                    "xmax": 250 * ureg.m,
                    "ymin": -250 * ureg.m,
                    "ymax": 250 * ureg.m,
                    "z": 6.75 * ureg.m,
                },
            }
        ],
        illumination={
            "type": "directional",
            "zenith": 40 * ureg.deg,
            "azimuth": 0.0 * ureg.deg,
            "irradiance": 20.0,
        },
        integrator={"type": "path", "max_depth": -1},
    )



from __future__ import annotations

import attrs
import numpy as np
import pint
import pinttr

from ._core import AbstractDirectionalIllumination
from ...attrs import documented, parse_docs
from ...frame import angles_to_direction
from ...units import unit_context_config as ucc
from ...units import unit_registry as ureg
from ...validators import is_positive


@parse_docs
@attrs.define(eq=False, slots=False)
class AstroObjectIllumination(AbstractDirectionalIllumination):
    """
    Astronomical Object Illumination scene element [``astro_object``].

    This illumination represents the light coming from a distant astronomical
    object (*e.g.* the Sun). The astronomical object uniformly illuminates a
    portion of the sky such that it appears to the observer as a circle. Its
    size is defined by the apparent diameter of the circle, expressed in
    degrees.

    Warnings
    --------
    This is an experimental feature. At the moment, using the
    :class:`.DirectionalIllumination` is recommended.

    Notes
    -----
    Contrary to the directional illuminant, the astronomical object is not a
    delta emitter, *i.e.* it has a non-zero apparent size. Light rays cast by
    this illuminant are, from the point of view of the observer, encompassed
    in a cone, while they are parallel in the case of the directional
    illumination.
    """

    angular_diameter: pint.Quantity = documented(
        pinttr.field(
            default=0.5358 * ureg.deg,
            validator=[is_positive, pinttr.validators.has_compatible_units],
            units=ucc.deferred("angle"),
        ),
        doc="Apparent diameter of the celestial body as seen from the point. "
        "The default value is an average of the apparent diameter of the "
        "Sun as seen from Earth.\n"
        "\n"
        "Unit-enabled field (default units: ucc['angle']).",
        type="quantity",
        init_type="quantity or float",
        default="0.5358 deg",
    )

    @property
    def direction(self) -> np.ndarray:
        """
        Illumination direction as an array of shape (3,), pointing inwards.
        """
        return angles_to_direction(
            [self.zenith.m_as(ureg.rad), self.azimuth.m_as(ureg.rad)],
            azimuth_convention=self.azimuth_convention,
            flip=False,
        ).reshape((3,))

    @property
    def template(self) -> dict:
        # Inherit docstring
        return {
            "type": "astroobject",
            "to_world": self._to_world,
            "angular_diameter": self.angular_diameter.m_as("degree"),
        }



from __future__ import annotations

import warnings

import attrs
import mitsuba as mi
import numpy as np
import pint
import pinttr

from ._core import ShapeInstance
from ..core import BoundingBox, traverse
from ...attrs import documented, parse_docs
from ...contexts import KernelContext
from ...kernel import UpdateParameter
from ...units import unit_context_config as ucc
from ...units import unit_context_kernel as uck


@parse_docs
@attrs.define(eq=False, slots=False)
class BufferMeshShape(ShapeInstance):
    """
    Buffer mesh shape [``buffer_mesh``].

    This shape represents a triangulated mesh directly defined by lists of vertex
    positions and faces.

    Notes
    -----
    * The `buffermesh` class does not support the `to_world` parameter. Mesh vertices
      must be transformed manually. The Eradiate documentation contains a tutorial.
    """

    vertices: pint.Quantity = documented(
        pinttr.field(
            validator=pinttr.validators.has_compatible_units,
            units=ucc.deferred("length"),
            kw_only=True,
        ),
        doc="List of vertex coordinates, specified either as a (n, 3) NumPy "
        "array or a list of triplets.\n"
        "\n"
        "Unit-enabled field (default: ucc['length']).",
        type="quantity",
        init_type="array-like",
    )

    faces: np.ndarray = documented(
        attrs.field(
            kw_only=True,
            converter=np.array,
        ),
        doc="List of face definitions. specified either as a (n, 3) NumPy "
        "array or a list of triplets of vertex indices.",
        type="ndarray",
        init_type="array-like",
    )

    @vertices.validator
    @faces.validator
    def _vertex_face_validator(self, attribute, value):
        if value.ndim != 2 or value.shape[0] == 0 or value.shape[1] != 3:
            raise ValueError(
                f"while validating {attribute.name}, must be an array of shape "
                f"(n, 3), got {value.shape}"
            )

    def __attrs_post_init__(self):
        if self.to_world is not None:
            warnings.warn(
                "A BufferMeshShape instance has its to_world field set: It will be ignored. "
                "For information on how to transform mesh vertices, see the Eradiate documentation."
            )

        self.update()

    def bbox(self) -> BoundingBox:
        # Inherit docstring
        raise NotImplementedError

    @property
    def instance(self) -> mi.Object:
        if self.bsdf is not None:
            template, _ = traverse(self.bsdf)
            kdict = template.render(ctx=KernelContext())
            kdict["id"] = self._bsdf_id  # TODO: Enforce ID control at BSDF level
            bsdf = mi.load_dict(kdict)
        else:
            bsdf = None

        props = mi.Properties()
        props["mesh_bsdf"] = bsdf

        mesh = mi.Mesh(
            name=self.id,
            face_count=self.faces.shape[0],
            vertex_count=self.vertices.shape[0],
            has_vertex_normals=False,
            has_vertex_texcoords=False,
            props=props,
        )

        vertices = self.vertices.m_as(uck.get("length"))
        mesh_params = mi.traverse(mesh)
        mesh_params["vertex_positions"] = vertices.ravel()
        mesh_params["faces"] = self.faces.ravel()
        mesh_params.update()

        return mesh

    def write_ply(self, filename: str) -> None:
        """
        Write the mesh data to a PLY file.

        Parameters
        ----------
        filename : str
            Path and filename to write the mesh file into. No directories
            are created.

        Notes
        -----
        Vertex coordinates are expressed in kernel units and accordingly prior
        to writing to disk. See the documentation of
        :data:`eradiate.unit_context_kernel` and the
        :ref:`sec-user_guide-unit_guide_user`.
        """
        self.instance.write_ply(filename)

    @property
    def params(self) -> dict[str, UpdateParameter] | None:
        if self.bsdf is None:
            return None

        _, params = traverse(attrs.evolve(self.bsdf, id=self._bsdf_id))
        return {f"bsdf.{k}": v for k, v in params.items()}



from __future__ import annotations

import enum
import importlib.resources
import typing as t
from pathlib import Path

from dynaconf import Dynaconf, Validator

from ._env import ENV, SOURCE_DIR
from ..frame import AzimuthConvention


class ProgressLevel(enum.IntEnum):
    """
    An enumeration defining valid progress levels.

    This is an integer enumeration, meaning that levels can be compared to
    numerics.
    """

    @staticmethod
    def convert(value: t.Any) -> ProgressLevel:
        """
        Attempt conversion of a value to an :class:`.ProgressLevel`
        instance. The conversion protocol is as follows:

        * If ``value`` is a string, it is converted to upper case and passed to
          the indexing operator of :class:`.ProgressLevel`.
        * If ``value`` is an integer, it is passed to the call operator of
          :class:`.ProgressLevel`.
        * If ``value`` is a :class:`.ProgressLevel` instance, it is returned
          without change.
        * Otherwise, the method raises an exception.

        Parameters
        ----------
        value
            Value to attempt conversion of.

        Returns
        -------
        Converted value

        Raises
        ------
        TypeError
            If no conversion protocol exists for ``value``.
        """
        if isinstance(value, ProgressLevel):
            return value
        elif isinstance(value, str):
            return ProgressLevel[value.upper()]
        elif isinstance(value, int):
            return ProgressLevel(value)
        else:
            raise TypeError(f"Cannot convert a {type(value)} instance to ProgressLevel")

    NONE = 0  #: No progress
    SPECTRAL_LOOP = enum.auto()  #: Up to spectral loop level progress
    KERNEL = enum.auto()  #: Up to kernel level progress


def _default_download_dir():
    return (
        Path("./eradiate_downloads").absolute().resolve()
        if SOURCE_DIR is None
        else SOURCE_DIR / ".eradiate_downloads"
    )


DEFAULTS = importlib.resources.files("eradiate.config").joinpath(f"{ENV}.toml")

#: Main settings data structure. See the `Dynaconf documentation <https://www.dynaconf.com/>`_
#: for details.
settings = Dynaconf(
    settings_files=[DEFAULTS, "eradiate.toml"],
    envvar_prefix="ERADIATE",
    merge_enabled=True,
    validate_on_update=True,
    validators=[
        Validator("AZIMUTH_CONVENTION", cast=AzimuthConvention.convert),
        Validator("DATA_STORE_URL", cast=str),
        Validator(
            "DOWNLOAD_DIR",
            default=_default_download_dir(),
            cast=lambda x: Path(x).resolve(),
        ),
        Validator("OFFLINE", cast=bool),
        Validator("PROGRESS", cast=ProgressLevel.convert),
        Validator("SMALL_FILES_REGISTRY_URL", cast=str),
        Validator("SMALL_FILES_REGISTRY_REVISION", cast=str),
    ],
)



from __future__ import annotations

from pathlib import Path

import attrs
import drjit as dr
import mitsuba as mi
import numpy as np
import pint
import pinttr

from ._core import Illumination
from ..core import NodeSceneElement
from ..spectra import Spectrum, spectrum_factory
from ... import validators
from ...attrs import documented, parse_docs
from ...units import unit_context_config as ucc
from ...units import unit_context_kernel as uck
from ...units import unit_registry as ureg
from ...validators import has_quantity


@parse_docs
@attrs.define(eq=False, slots=False)
class SpotIllumination(Illumination):
    """
    Spot illumination scene element [``spot``].

    Eradiate ships a beam texture that implements a gaussian beam profile with
    three standard deviations included in the total beam width. This texture is
    named ``gaussian_3sigma.bmp`` and can be used by retrieving the path to the
    file using `eradiate.data.data_store.fetch()`.
    """

    origin: pint.Quantity = documented(
        pinttr.field(
            factory=lambda: [1, 1, 1] * ureg.m,
            validator=[validators.has_len(3), pinttr.validators.has_compatible_units],
            units=ucc.deferred("length"),
        ),
        doc="A 3-vector specifying the position of the spot.\n"
        "\n"
        "Unit-enabled field (default: ucc['length']).",
        type="quantity",
        init_type="array-like",
        default="[1, 1, 1] m",
    )

    target: pint.Quantity = documented(
        pinttr.field(
            factory=lambda: [0, 0, 0] * ureg.m,
            validator=[validators.has_len(3), pinttr.validators.has_compatible_units],
            units=ucc.deferred("length"),
        ),
        doc="Point location targeted by the spot.\n"
        "\n"
        "Unit-enabled field (default: ucc['length']).",
        type="quantity",
        init_type="array-like",
        default="[0, 0, 0] m",
    )

    @target.validator
    @origin.validator
    def _target_origin_validator(self, attribute, value):
        if np.allclose(self.target, self.origin):
            raise ValueError(
                f"While initializing {attribute}: "
                f"Origin and target must not be equal, "
                f"got target = {self.target}, origin = {self.origin}"
            )

    up: np.ndarray = documented(
        attrs.field(
            converter=np.array,
            validator=validators.has_len(3),
        ),
        doc="A 3-vector specifying the up direction of the spot.\n"
        "This vector must be different from the spots's pointing direction,\n"
        "which is given by ``target - origin``.",
        type="array",
        init_type="array-like",
        default="[0, 0, 1]",
    )

    @up.default
    def _up_factory(self):
        direction = dr.normalize(
            mi.ScalarVector3f(
                self.origin.m_as(ucc.get("length"))
                - self.target.m_as(ucc.get("length"))
            )
        )
        return mi.coordinate_system(direction)[0]

    @up.validator
    def _up_validator(self, attribute, value):
        direction = self.target - self.origin
        if np.allclose(np.cross(direction, value), 0):
            raise ValueError(
                f"While initializing '{attribute.name}': "
                f"up direction must not be colinear with viewing direction, "
                f"got up = {self.up}, direction = {direction}"
            )

    beam_width: pint.Quantity = documented(
        pinttr.field(default=10.0 * ureg.deg, units=ucc.deferred("angle")),
        doc="Spot light beam width.\n\nUnit-enabled field (default: degree).",
        type="quantity",
        init_type="quantity or float",
        default="10°",
    )

    intensity: Spectrum = documented(
        attrs.field(
            default=1.0,
            converter=spectrum_factory.converter("intensity"),
            validator=[
                attrs.validators.instance_of(Spectrum),
                has_quantity("intensity"),
            ],
        ),
        doc="Emitted power in the plane orthogonal to the illumination direction. "
        "Must be an intensity spectrum (in W/sr/nm or compatible unit). "
        "Can be initialised with a dictionary processed by "
        ":meth:`.SpectrumFactory.convert`.",
        type=":class:`~eradiate.scenes.spectra.Spectrum`",
        init_type=":class:`~eradiate.scenes.spectra.Spectrum` or dict or float",
        default="1.0 ucc[intensity]",
    )

    beam_profile: Path | None = documented(
        attrs.field(
            default=None,
            converter=attrs.converters.optional(Path),
        ),
        type="Path",
        init_type="path-like",
        default=None,
        doc="Path to the file describing the beam profile. Must be a valid bitmap "
        "image file.",
    )

    @classmethod
    def from_size_at_target(
        cls,
        target: np.typing.ArrayLike,
        direction: np.typing.ArrayLike,
        spot_radius: pint.Quantity,
        beam_width: pint.Quantity,
        **kwargs,
    ) -> SpotIllumination:
        """
        Create a :class:`.SpotIllumination` which illuminates a region of a
        specified size around the target point.

        The illuminated area is a circle defined by the spot radius around the
        target point in the plane orthogonal to the spot direction. The spot
        origin is inferred from the spot radius and a beam width value
        (in angle units): the spot origin will be positioned closer to the
        target point as the beam width increases and further from the target
        point as the beam width decreases.

        Parameters
        ----------
        target : array-like
            Target point for the spot.
            Unitless values are converted to ``ucc['length']``.

        direction : array-like
            Pointing direction for the spot, towards target.

        spot_radius : float or quantity
            Radius of the desired spot at the target position.
            Unitless values are converted to ``ucc['length']``.

        beam_width : float or quantity
            Divergence angle of the spot.
            Unitless values are converted to ``ucc['angle']``.

        **kwargs
            Remaining keyword arguments are forwarded to the
            :class:`.SpotIllumination` constructor.

        Returns
        -------
        SpotIllumination
        """

        if "origin" in kwargs:
            raise TypeError(
                "The from_size_at_target constructor computes the origin position."
            )

        target = pinttr.util.ensure_units(target, default_units=ucc.get("length"))
        half_angle = (
            pinttr.util.ensure_units(beam_width, default_units=ureg.radian) / 2.0
        )
        tan_divergence = np.tan(half_angle)
        distance = spot_radius / tan_divergence

        origin = target - (direction * distance)

        return cls(origin=origin, target=target, beam_width=beam_width, **kwargs)

    @property
    def _to_world(self) -> mi.ScalarTransform4f:
        target = self.target.m_as(uck.get("length"))
        origin = self.origin.m_as(uck.get("length"))
        return mi.ScalarTransform4f.look_at(origin=origin, target=target, up=self.up)

    @property
    def template(self) -> dict:
        retdict = {
            "type": "spot",
            "beam_width": self.beam_width.m_as(uck.get("angle")),
            "cutoff_angle": self.beam_width.m_as(uck.get("angle")),
            "to_world": self._to_world,
        }

        if self.beam_profile is not None:
            retdict["texture"] = {
                "type": "bitmap",
                "filename": str(self.beam_profile),
            }
        return retdict

    @property
    def objects(self) -> dict[str, NodeSceneElement]:
        return {"intensity": self.intensity}



import numpy as np
import xarray as xr

from .interp import dataarray_to_rgb, film_to_angular


@xr.register_dataarray_accessor("ert")
class EradiateDataArrayAccessor:
    """
    Convenience wrapper for operations on :class:`~xarray.DataArray`
    instances. Accessed as a ``DataArray.ert`` property.
    """

    def __init__(self, xarray_obj):
        self._obj = xarray_obj

    def to_angular(
        self,
        theta: np.typing.ArrayLike,
        phi: np.typing.ArrayLike,
        x_label: str = "x",
        y_label: str = "y",
        theta_label: str = "theta",
        phi_label: str = "phi",
    ) -> xr.DataArray:
        """
        Attempt interpolation of self to angular coordinates using
        :func:`~eradiate.xarray.interp.film_to_angular`.
        """
        return film_to_angular(
            self._obj,
            theta=theta,
            phi=phi,
            x_label=x_label,
            y_label=y_label,
            theta_label=theta_label,
            phi_label=phi_label,
        )

    def to_rgb(
        self,
        channels,
        normalize: bool = True,
        gamma_correction: bool = True,
    ) -> np.ndarray:
        """
        Generate a basic RGB image as a Numpy array from self using
        :func:`~eradiate.xarray.interp.dataarray_to_rgb`.
        """
        return dataarray_to_rgb(
            self._obj,
            channels,
            normalize=normalize,
            gamma_correction=gamma_correction,
        )



from __future__ import annotations

from abc import ABC

import attrs
import mitsuba as mi

from ..bsdfs import BSDF, bsdf_factory
from ..core import BoundingBox, InstanceSceneElement, NodeSceneElement, Ref
from ... import converters
from ..._factory import Factory
from ...attrs import documented, get_doc, parse_docs

shape_factory = Factory()
shape_factory.register_lazy_batch(
    [
        ("_cuboid.CuboidShape", "cuboid", {}),
        ("_rectangle.RectangleShape", "rectangle", {}),
        ("_sphere.SphereShape", "sphere", {}),
        ("_filemesh.FileMeshShape", "file_mesh", {}),
        ("_buffermesh.BufferMeshShape", "buffer_mesh", {}),
    ],
    cls_prefix="eradiate.scenes.shapes",
)


@parse_docs
@attrs.define(eq=False, slots=False)
class Shape:
    """
    Abstract interface for all shape scene elements.

    Notes
    -----
    * This class is to be used as a mixin.
    """

    id: str | None = documented(
        attrs.field(
            default="shape",
            validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        ),
        doc=get_doc(NodeSceneElement, "id", "doc"),
        type=get_doc(NodeSceneElement, "id", "type"),
        init_type=get_doc(NodeSceneElement, "id", "init_type"),
        default='"shape"',
    )

    bsdf: BSDF | Ref | None = documented(
        attrs.field(
            default=None,
            converter=attrs.converters.optional(bsdf_factory.convert),
            validator=attrs.validators.optional(
                attrs.validators.instance_of((BSDF, Ref))
            ),
        ),
        doc="BSDF attached to the shape. If a dictionary is passed, it is "
        "interpreted by :class:`bsdf_factory.convert() <.Factory>`. "
        "If unset, no BSDF will be specified during the kernel dictionary "
        "generation: the kernel's default will be used. If a :class:`.BSDF` "
        "instance (or a corresponding dictionary specification) is passed, "
        "its `id` member is automatically overridden.",
        type="BSDF or Ref or None",
        init_type="BSDF or Ref or dict, optional",
    )

    to_world: "mitsuba.ScalarTransform4f" = documented(
        attrs.field(
            converter=converters.to_mi_scalar_transform,
            default=None,
        ),
        doc="Transform to scale, shift and rotate the shape. ",
        type="mitsuba.ScalarTransform4f or None",
        init_type="mitsuba.ScalarTransform4f or array-like, optional",
        default=None,
    )

    @to_world.validator
    def to_world_validator(self, attribute, value):
        if value is not None:
            if not isinstance(value, mi.ScalarTransform4f):
                raise TypeError(
                    f"while validating '{attribute.name}': "
                    f"'{attribute.name}' must be a mitsuba.ScalarTransform4f; "
                    f"found: {type(value)}",
                )

    def __attrs_post_init__(self):
        self.update()

    def update(self) -> None:
        # Inherit docstring

        # Normalize child BSDF ID
        if isinstance(self.bsdf, BSDF):
            self.bsdf.id = self._bsdf_id

    @property
    def bbox(self) -> BoundingBox:
        """
        :class:`.BoundingBox` : Shape bounding box. Default implementation
            raises a :class:`NotImplementedError`.
        """
        raise NotImplementedError

    @property
    def _bsdf_id(self) -> str:
        return f"{self.id}_bsdf"

    @property
    def objects(self) -> dict[str, NodeSceneElement] | None:
        # Inherit docstring
        if self.bsdf is None:
            return None
        else:
            return {"bsdf": self.bsdf}


@attrs.define(eq=False, slots=False)
class ShapeNode(Shape, NodeSceneElement, ABC):
    """
    Interface for shapes which can be represented as Mitsuba scene dictionary
    nodes.
    """

    pass


@attrs.define(eq=False, slots=False)
class ShapeInstance(Shape, InstanceSceneElement, ABC):
    """
    Interface for shapes which have to be expanded as Mitsuba objects.
    """

    pass



"""
This module contains definitions for all post-processing pipeline operations.
The implementation complies with the `Hamilton <https://hamilton.dagworks.io/>`_
dataflow framework's specifications: the post-processing pipeline is modelled as
a directed acyclic graph (DAG), and all post-processing operations are a node in
that DAG.

Eradiate uses advanced Hamilton features to mutate the pipeline based on
configuration input. For example:

* Node names reflect the processed physical quantity.
* Viewing angles are only computed if the measure is distant.
* Post-processing operations vary depending on the active mode. For example,
  SRF weighted averaging is only performed in CKD modes.

For clarity, the implementation is split as follows:

* :mod:`eradiate.pipelines.logic` provides the fundamental logic powering each
  step of post-processing;
* :mod:`eradiate.pipelines.definitions` contains the pipeline definition and is
  intended to be consumed by the Hamilton driver constructor for pipeline
  initialization;
* :mod:`eradiate.pipelines.core` provides convenience entry points to facilitate
  post-processing pipeline initialization.

See Also
--------

* :class:`hamilton.driver.Driver`
* `Hamilton's documentation on tags \
  <https://hamilton.dagworks.io/en/latest/reference/decorators/tag/#tag>`_
"""

import lazy_loader

__getattr__, __dir__, __all__ = lazy_loader.attach_stub(__name__, __file__)

del lazy_loader
